import numpy as np
import pandas as pd
from abc import ABC
from abc import abstractmethod

class WeightUpdate(ABC):
    """
        Interface for updating the FCM weight matrix
    """
    @abstractmethod
    def update(**kwargs) -> np.ndarray:
        """
            Update the FCM weight matrix
        """
        raise NotImplementedError('Update method is not defined!')


class NhlWeightUpdate(WeightUpdate):
    """
        Synchronously (i.e., NHL) update the weight matrix according 
        to the adjusted Hebbian rule.
    """
    @staticmethod
    def update(**kwargs) -> pd.DataFrame:
        """
            Update the weight matrix according to the adjusted Hebbian rule.

            Parameters
            ----------
            initial_state: dict
                            initial state vector of the concepts

            weight_matrix: pd.DataFrame
                            N*N weight matrix of the FCM
            
            gamma: float, int
                    decay coefficient
            
            eta: float,
                    learning rate
        
            Return
            ----------
            y: pd.DataFrame
                updated weight matrix
        """
        # Extract the arguments
        initial_state = kwargs['initial_state'].copy()
        weight_matrix = kwargs['weight_matrix'].copy()
        gamma = kwargs['gamma']
        eta = kwargs['eta']

        # convert pd -> ndarray
        # convert dict val -> array
        w = weight_matrix.to_numpy()
        s = np.array(list(initial_state.values()))

        # update the W matrix
        a = gamma*w
        b = eta*s
        c = abs(w)*s
        res = a + abs(np.sign(w))*b*(s-c.T).T
        
        # convert it back to df
        res = pd.DataFrame(res, columns = weight_matrix.columns, index = weight_matrix.columns)
        return res


class AhlWeightUpdate(WeightUpdate):
    """
        Asynchronous (i.e., AHL) Update the weight matrix according to the adjusted Hebbian rule.
    """
    @staticmethod
    def update(**kwargs) -> float:
        """
            Asynchronously update the weight matrix according to the adjusted Hebbian rule.

            Parameters
            ----------
            source: str,
                        the activated concept

            state_vector: dict
                            state vector of the concepts
                            keys ---> concepts, values ---> value of the associated concept

            weight_matrix: pd.DataFrame
                            N*N weight matrix of the FCM.

            gamma: float, int
                    decay coefficient
            
            eta: float,
                    learning rate           
            
            Return
            ------
            y: pd.DataFrame
                updated weight matrix
        """
        # Extract arguments
        source = kwargs['source']
        state_vector = kwargs['state_vector']
        weight_matrix = kwargs['weight_matrix']
        gamma = kwargs['gamma']
        eta = kwargs['eta']

        gamma = 1-gamma
        w = weight_matrix.copy()
        for j in source:
            for i in weight_matrix:
                if j != i:
                    w.loc[j, i] = gamma*w.loc[j, i]+eta*state_vector[j]*(state_vector[i]-w.loc[j, i]*state_vector[j])
        return w
        


import pandas as pd
import warnings
from typing import Union
from tqdm import tqdm
from fcmpy.expert_fcm.input_validator import type_check
from .nhl import HebbianLearning
from .update_weight import AhlWeightUpdate
from .update_state import FcmUpdateAsynch
from .termination import FirstCriterion
from .termination import SecondCriterion
from .update_parameters import Eta
from .update_parameters import Gamma


class AHL(HebbianLearning):
    """
        AHL algorithm for FCMs.
    """
    @type_check
    def __init__(self,  state_vector: dict, weight_matrix: pd.DataFrame,
                    activation_pattern:dict,  doc_values:dict) -> None:
        """
            Active Hebbian Learning for FCMs.

            Parameters
            ----------
            state_vector: dict
                            state vector of the concepts
                            keys ---> concepts, values ---> value of the associated concept

            weight_matrix: pd.DataFrame
                                N*N weight matrix of the FCM.

            activation_pattern: dict
                                    sequence of activation concepts.
                                    keys ---> sequence step, values --> activated concept.

            doc_values: dict
                            Desired Output Concepts (DOCs) values.
                            keys ---> output concepts, values ---> desired output range ([min, max]).
                            e.g.,
                            doc_values = {'C1':[0.68,0.74], 'C5':[0.74,0.8]}
        """
        self.__state_vector = state_vector
        self.__weight_matrix = weight_matrix
        self.__doc_values = doc_values
        self.__activation_pattern = activation_pattern
        self.__weightUpdate = AhlWeightUpdate()
        self.__stateUpdate = FcmUpdateAsynch()
        self.__termination1 = FirstCriterion()
        self.__termination2 = SecondCriterion()
        self.__update_eta = Eta()
        self.__update_gamma = Gamma()
    
    @type_check
    def run(self, decay=0.03, learning_rate=0.01, iterations:int= 100, transfer:str= 'sigmoid', 
                    thresh:float = 0.002, l:Union[float, int]=0.98, auto_learn=False, **kwargs) -> pd.DataFrame:
        
        """
            Run the AHL algorithm for FCMs

            Parameters
            ----------
            decay: float
                    default --> 0.03
            
            learning_rate: float
                            default --> 0.01
            
            iterations: int
                        default --> 100
            
            transfer: str
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"
                        default --> "sigmoid"
            
            thresh: float
                        default --> 0.002
            
            l: Union[float, int]
                    default --> 0.98
            
            auto_learn: bool
                        default --> False

            Return
            ------
            y: pd.DataFrame
                the optimized weight matrix
        """
        w_prior = self.__weight_matrix.copy()
        s_prior = self.__state_vector.copy()
        
        for _ in tqdm(range(iterations)):
            for cycle in self.__activation_pattern:
                
                if auto_learn:
                    learning_rate = self.__update_eta.update(cycle=(cycle+1), b1=kwargs['b1'], l1=kwargs['lbd1']) 
                    decay = self.__update_gamma.update(cycle=(cycle+1), b2=kwargs['b2'], l2=kwargs['lbd2']) 

                # Update state
                s_new = self.__stateUpdate.update(source=self.__activation_pattern[cycle], state_vector=s_prior,
                                                    weight_matrix=w_prior, transfer=transfer, l = l)
                # Update weights
                w_new = self.__weightUpdate.update(source=self.__activation_pattern[cycle], state_vector=s_new, 
                                                    weight_matrix=w_prior, eta=learning_rate, gamma=decay)
                # check termination
                if self.__termination1.terminate(doc_values=self.__doc_values, state_vector_prior=s_prior, state_vector_current=s_new) and \
                                                    self.__termination2.terminate(doc_values=self.__doc_values, state_vector_prior=s_prior, 
                                                                        state_vector_current=s_new, thresh=thresh):
                
                    print(f'The AHL learning process converged at step {_} with the learning rate eta = {learning_rate} and decay = {decay}!')
                    return w_new                   
                else:
                    w_prior = w_new
                    s_prior = s_new
        if _ >= iterations-1:
            warnings.warn(f'The AHL did not converge with the learning rate learning rate: {learning_rate} and decay: {decay}! Consider a different set of parameters.')
            return w_new


import pandas as pd
import numpy as np
import warnings
from typing import Union
from abc import ABC, abstractmethod
from fcmpy.store.methodsStore import InferenceStore
from fcmpy.store.methodsStore import TransferStore
from fcmpy.store.methodsStore import ConvergenceStore
from fcmpy.expert_fcm.input_validator import type_check


class Simulator(ABC):
    """
        Class of methods for simulating FCMs.
    """
    @abstractmethod
    def simulate():
        raise NotImplementedError('Simulate method is not defined!')


class FcmSimulator(Simulator):
    """
        The class includes methods for running simulations on top of a defined FCM.

        Methods:
            simulate(initial_state: dict, weight_matrix: Union[pd.DataFrame, np.ndarray], 
                            transfer: str, inference: str, thresh:float=0.001, iterations:int=50, 
                            output_concepts = None, convergence = 'absDiff', **kwargs)
    """
    @staticmethod
    @type_check
    def __getStableConcepts(weight_matrix: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        """
            Extract the positions of the stable concepts (concepts with in-degree == 0).

            Parameters
            ----------
            weight_matrix: numpy.ndarray
                            N*N weight matrix of the FCM.
            
            Return
            ----------
            y: numpy.ndarray
                    the positions of the stable concepts (concepts with in-degree == 0)
        """
        stables = []
        for i in range(len(weight_matrix)):
            if np.all(weight_matrix[i] == 0):
                stables.append(i)

        return stables
    
    @staticmethod
    @type_check
    def __simInputTransform(weight_matrix: Union[pd.DataFrame, np.ndarray], initial_state: dict) -> tuple:
        """
            Transform the weight matrix and initial state vector for the simulation.
        """
        if type(weight_matrix) != np.ndarray:
            # Align the initial_vector order for correct computations (vec . mat)
            initial_state = {k : initial_state[k] for k in weight_matrix.columns}
            weight_matrix=weight_matrix.to_numpy()
        else:
            warnings.warn("When passing an initial state with a weight matrix type \
                numpy.ndarray make sure that the order of the keys in the dictionary \
                with the initial states matches the order of the column of the numpy.ndarray!")
        
        return weight_matrix, initial_state

    @staticmethod
    @type_check
    def __outputConceptCheck(output_concepts: Union[None, list], initial_state: dict):
        """
            Check if output concepts are in the initial_state.keys()
        """
        if output_concepts:
            r = set(output_concepts) - set(initial_state.keys())
            if r:
                raise ValueError(f'The specified output concept {r} is not in the list.')
    
    @staticmethod
    @type_check
    def simulate(initial_state: dict, weight_matrix: Union[pd.DataFrame, np.ndarray], 
                        transfer: str, inference: str, thresh:float=0.001, iterations:int=50, 
                        output_concepts = None, convergence = 'absDiff', **kwargs) -> pd.DataFrame:
        """
            Runs simulations over the passed FCM.
            
            Parameters
            ----------
            initial_state: dict
                            initial state vector of the concepts
                            keys ---> concepts, values ---> initial state of the associated concept

            weight_matrix: pd.DataFrame, np.ndarray
                            N*N weight matrix of the FCM.

            transfer: str
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"

            inference: str
                        inference method --> "kosko", "mKosko", "rescaled"

            thresh: float
                        threshold for the error

            iterations: int
                            number of iterations

            output_concepts: bool, list
                                the output concepts for the convergence check
                                default --> None

            convergence: str,
                            convergence method
                            default --> 'absDiff': absolute difference between the simulation steps

            kwargs: additional parameters for the methods (e.g., lambda (l) parameter for the sigmoid function)

            Return
            ----------
            y: pandas.DataFrame
                    results of the simulation.
        """
        # Check the output concepts.
        FcmSimulator.__outputConceptCheck(output_concepts=output_concepts,
                                        initial_state=initial_state)

        # Transform the inputs if necessary.
        weight_matrix, initial_state = FcmSimulator.__simInputTransform(weight_matrix=weight_matrix,
                                                                        initial_state=initial_state)

        # create the empty dataframe for the results
        results = pd.DataFrame(initial_state, index=[0])
        state_vector = np.array(list(initial_state.values()))

        # get the stable concept values
        stableConceptPos = FcmSimulator.__getStableConcepts(weight_matrix=weight_matrix.T)
        satble_values = state_vector[stableConceptPos]

        # get the methods for the simulation.
        transfer = TransferStore.get(transfer)()
        inference = InferenceStore.get(inference)()
        conv = ConvergenceStore.get(convergence)()
        
        # initialize params
        convergenceStatus = False
        step_count = 0

        for _ in range(iterations):
            if not convergenceStatus:    
                infered = inference.infer(initial_state=state_vector, weight_matrix=weight_matrix, params=kwargs) # Inference
                state_vector = transfer.transfer(x=infered, params=kwargs) # Apply transfer func on the results
                
                # Reset the stable values
                state_vector[stableConceptPos] = satble_values

                # Append the results
                results.loc[len(results)] = state_vector

                # update the step_count
                step_count +=1
                
                # compute the residuals between the steps.
                convergenceStatus = conv.check_convergence(output_concepts=output_concepts, results=results, threshold = thresh, params=kwargs) 
            else:    
                print(f'The values converged in the {step_count+1} state (e <= {thresh})')
                break

            if step_count >= iterations:
                warnings.warn("The values didn't converge. More iterations are required!")

        return results


import pandas as pd
import numpy as np
from abc import ABC, abstractmethod

class InterventionConstructor(ABC):
    """
        Class of methods for building interventions.
    """
    @abstractmethod
    def build() -> dict:
        raise NotImplementedError('Build method is not defined!')


class SingleShot(InterventionConstructor):
    """
        Construct single-shot interventions
    """ 
    @staticmethod
    def build(**kwargs) -> dict:
        """
            Construct a single shot intervention case

            Parameters
            ----------
            state_vector: dict
                            keys --> concepts, values --> state of the concept.

            weight_matrix: pd.DataFrame
                            causal weights between concepts

            Return
            ----------
            y: dictionary
        """
        intervention = {}
        intervention['state_vector'] = kwargs['initial_state']
        intervention['weight_matrix'] = kwargs['weight_matrix']
        
        return intervention


class Continuous(InterventionConstructor):
    
    @staticmethod
    def build(**kwargs) -> dict:
        """
            Construct a continuous intervention case

            Parameters
            ----------
            weight_matrix: pd.DataFrame
                            causal weights between concepts

            initial_state: dict
                            keys --> concepts, values --> state of the concept.

            equilibriums: dict
                            keys --> concepts, values --> equilibrium states of the concept.

            impact: dict
                        keys ---> concepts the intervention impacts, value: the associated causal weight

            effectiveness: float
                            the degree to which the intervention was delivered (should be between [-1, 1])
                            default --> 1
                            
            Return
            ----------
            y: dictionary
        """
        weight_matrix = kwargs['weight_matrix']
        initial_state = kwargs['initial_state']
        equilibriums = kwargs['equilibriums']
        impact = kwargs['params']['impact']
        
        # Set the intervention effectiveness to 1 if the optional parameter is not specified.
        try:
            effectiveness = kwargs['params']['effectiveness']
        except:
            effectiveness = 1

        # Check whether the passed intervention inputs are in the function's domain.
        if (min(list(impact.values())) < -1) or (max(list(impact.values())) > 1):
            raise ValueError('the values in the causal weights are out of the domain [-1,1].')
        elif (effectiveness < 0) or (effectiveness > 1):
            raise ValueError('the values in the intervention effectiveness are out of the domain [0,1].')

        intervention = {}
        intervention['effectiveness'] = effectiveness
        
        # construct a weight matrix for a given intervention
        if type(weight_matrix) == np.ndarray:
            temp = pd.DataFrame(weight_matrix, columns=initial_state)
        else:
            temp = weight_matrix.copy(deep=True)

        temp['antecedent'] = temp.columns
        temp.set_index('antecedent', inplace=True)
        temp['intervention'] = 0
        temp.loc[len(temp)] = 0
        temp.rename(index = {temp.index[-1] : 'intervention'}, inplace = True)
        
        # add the intervention impact
        for key in impact.keys():
            temp.loc['intervention', key] = impact[key]
            
        # construct the new state vector for a given intervention (baseline + intervention effectiveness)
        temp_vector = equilibriums['baseline'].copy(deep=True)
        temp_vector = temp_vector.append(pd.Series({'intervention': effectiveness})).to_dict()
        
        # add the causal weights for the intervention
        intervention['weight_matrix'] = temp
        intervention['state_vector'] = temp_vector

        return intervention


from abc import ABC, abstractclassmethod 
import skfuzzy as fuzz
from fcmpy.expert_fcm.input_validator import type_check


class Defuzzification(ABC):
    """
        Defuzzification methods.
    """
    @abstractclassmethod
    def defuzz() -> float:
        raise NotImplementedError('defuzzification method is not defined!')


class Centroid(Defuzzification):
    """
        Centroid difuzzification method (i.e., center of gravity).
    """
    @staticmethod
    @type_check
    def defuzz(**kwargs) -> float:
        """
            Centroid difuzzification method (i.e., center of gravity).
            
            Other Parameters
            ----------
            **x: numpy.ndarray
                universe of discourse 
            
            **mfx: numpy.ndarray,
                        "aggregated" membership functions
            
            Return
            -------
            y: float
                defuzzified value
        """
        method = kwargs['method']
        x = kwargs['x']
        mfx = kwargs['mfx']

        return fuzz.defuzz(x, mfx, method)


class Bisector(Defuzzification):
    """
        Bisector difuzzification method.
    """
    @staticmethod
    @type_check
    def defuzz(**kwargs) -> float:
        """
            Bisector difuzzification method.
            
            Other Parameters
            ----------
            **x: numpy.ndarray
                universe of discourse 
            
            **mfx: numpy.ndarray,
                        "aggregated" membership functions
            
            Return
            -------
            y: float
                defuzzified value
        """
        method = kwargs['method']
        x = kwargs['x']
        mfx = kwargs['mfx']

        return fuzz.defuzz(x, mfx, method)


class MeanOfMax(Defuzzification):
    """
        MeanOfMax difuzzification method.
    """
    @staticmethod
    @type_check
    def defuzz(**kwargs) -> float:
        """
            MeanOfMax difuzzification method.
            
            Other Parameters
            ----------
            **x: numpy.ndarray
                universe of discourse 
            
            **mfx: numpy.ndarray,
                        "aggregated" membership functions
            
            Return
            -------
            y: float
                defuzzified value
        """
        method = kwargs['method']
        x = kwargs['x']
        mfx = kwargs['mfx']

        return fuzz.defuzz(x, mfx, method)


class MinOfMax(Defuzzification):
    """
        MinOfMax difuzzification method.
    """
    @staticmethod
    @type_check
    def defuzz(**kwargs) -> float:
        """
            MinOfMax difuzzification method.
            
            Other Parameters
            ----------
            **x: numpy.ndarray
                universe of discourse 
            
            **mfx: numpy.ndarray,
                        "aggregated" membership functions
            
            Return
            -------
            y: float
                defuzzified value
        """
        method = kwargs['method']
        x = kwargs['x']
        mfx = kwargs['mfx']

        return fuzz.defuzz(x, mfx, method)


class MaxOfMax(Defuzzification):
    """
        MaxOfMax difuzzification method.
    """
    @staticmethod
    @type_check
    def defuzz(**kwargs) -> float:
        """
            MaxOfMax difuzzification method.
            
            Other Parameters
            ----------
            **x: numpy.ndarray
                universe of discourse 
            
            **mfx: numpy.ndarray,
                        "aggregated" membership functions
            
            Return
            -------
            y: float
                defuzzified value
        """
        method = kwargs['method']
        x = kwargs['x']
        mfx = kwargs['mfx']

        return fuzz.defuzz(x, mfx, method)


import unittest
from fcmpy.simulator.simulator import FcmSimulator
from fcmpy.intervention.intervention import FcmIntervention
import pandas as pd

class TestIntervention(unittest.TestCase):
    
    def setUp(self):

        C1 = [0.0, 0.0, 0.6, 0.9, 0.0, 0.0, 0.0, 0.8]
        C2 = [0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.5]
        C3 = [0.0, 0.7, 0.0, 0.0, 0.9, 0.0, 0.4, 0.1]
        C4 = [0.4, 0.0, 0.0, 0.0, 0.0, 0.9, 0.0, 0.0]
        C5 = [0.0, 0.0, 0.0, 0.0, 0.0, -0.9, 0.0, 0.3]
        C6 = [-0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        C7 = [0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.4, 0.9]
        C8 =[0.1, 0.0, 0.0, 0.0, 0.0, 0.1, 0.6, 0.0]

        weight_matrix = pd.DataFrame([C1,C2, C3, C4, C5, C6, C7, C8], 
                            columns=['C1','C2','C3','C4','C5','C6','C7','C8'])

        init_state = {'C1': 1, 'C2': 1, 'C3': 0, 'C4': 0, 'C5': 0,
                            'C6': 0, 'C7': 0, 'C8': 0}
        
        self.inter = FcmIntervention(FcmSimulator)
        self.inter.initialize(initial_state=init_state, weight_matrix=weight_matrix, transfer='sigmoid', inference='mKosko', thresh=0.001, iterations=50, l=1)

    def test_addIntervention(self):
        # Check if the interventions were properly added to the constructor.
        self.inter.add_intervention('intervention_1', impact={'C1':-.3, 'C2' : .5}, effectiveness=1)
        self.inter.add_intervention('intervention_2', impact={'C1':-.5}, effectiveness=1)
        self.inter.add_intervention('intervention_3', impact={'C1':-1}, effectiveness=1)

        intervations = ['intervention_1', 'intervention_2', 'intervention_3']
        nInter = len(set(intervations) ^ set(self.inter.interventions.keys()))
        self.assertEqual(nInter, 0, msg="The interventions were not added properly to the constructor!")

    def test_removeIntervention(self):
        # Check if the interventions were properly added to the constructor.
        self.inter.add_intervention('intervention_1', impact={'C1':-.3, 'C2' : .5}, effectiveness=1)
        self.inter.add_intervention('intervention_2', impact={'C4':-.5}, effectiveness=1)
        self.inter.add_intervention('intervention_3', impact={'C5':-1}, effectiveness=1)

        self.inter.remove_intervention('intervention_1')
        intervations = ['intervention_2', 'intervention_3']
        nInter = len(set(intervations) ^ set(self.inter.interventions.keys()))
        self.assertEqual(nInter, 0, msg="The intervention was not removed from the constructor!")

    def test_testIntervention(self):
        # Check if the test_intervention runs properly.
        # Check if the stable concept (intervetion in this case) is indeed stable.
        self.inter.add_intervention('intervention_1', impact={'C1':-.3, 'C2' : .5})
        self.inter.add_intervention('intervention_2', impact={'C4':-.5})
        self.inter.add_intervention('intervention_3', impact={'C5':-1})

        self.inter.test_intervention('intervention_1')
        self.inter.test_intervention('intervention_2')
        self.inter.test_intervention('intervention_3')
        
        self.assertEqual(len(set(self.inter.test_results['intervention_1']['intervention'])), 1)
    
    def test_singleShot(self):
        self.inter.add_intervention('intervention_1', type='single_shot', initial_state = {'C1': 0.9, 'C2' : 0.4})
        self.assertEqual(self.inter.interventions['intervention_1']['state_vector']['C1'], 0.9)
        self.assertEqual(self.inter.interventions['intervention_1']['state_vector']['C2'], 0.4)
        
if __name__ == '__main__':
    unittest.main()


# Taken and modified from Simon Pirschel at
# https://aboutsimon.com/blog/2018/04/04/Python3-Type-Checking-And-Data-Validation-With-Type-Hints.html

from functools import wraps
from typing import get_type_hints, get_args
from inspect import getfullargspec


def validate_input(obj, **kwargs):
    """
    Check whether the passed arguments are of the same type
    as the pre-specified typehints.    
    """
    hints = get_type_hints(obj)

    # iterate all type hints
    for attr_name, attr_type in hints.items():
        if attr_name == 'return':
            continue
        if attr_name not in kwargs.keys(): # for default arguments (self)
            continue
        else:
            try:
                if not isinstance(kwargs[attr_name], attr_type):
                    raise TypeError(
                        'Argument %r is not of type %s' % (attr_name, attr_type)
                    )
            except: # for Union cases.
                if not isinstance(kwargs[attr_name], get_args(attr_type)):
                    raise TypeError(
                        'Argument %r is not of type %s' % (attr_name, attr_type)
                    )
                
def type_check(decorator):
    """
    Decorator for validating the typehints.
    """
    @wraps(decorator)
    def wrapped_decorator(*args, **kwargs):
        # translate *args into **kwargs
        func_args = getfullargspec(decorator)[0]
        kwargs.update(dict(zip(func_args, args)))

        validate_input(decorator, **kwargs)
        return decorator(**kwargs)

    return wrapped_decorator


import numpy as np 
import pandas as pd 
from abc import ABC, abstractclassmethod
from fcmpy.expert_fcm.input_validator import type_check
from fcmpy.expert_fcm.transform import Transform


class Entropy(ABC):
    """
        Entropy of the expert inputs.
    """
    @abstractclassmethod
    def calculateEntropy(data: pd.DataFrame, activationParamter):
        raise NotImplementedError('calculateEntropy method is not defined')

    
class InformationEntropy(Entropy):
    """
        Methods for calculating the information entropy.
    """
    @staticmethod
    @type_check
    def calculateEntropy(**kwargs) -> pd.DataFrame:
        """
            Calculate the information entropy of the expert ratings.

            Other Parameters
            ----------------
            **data: collections.OrderedDict
                    ordered dictionary with the expert inputs
            
            Return
            -------
            y: pandas.DataFrame,
                entropy of the concept pairs in expert ratings
        """
        data = kwargs['data']
        nExperts = len(data)
        flat_data = Transform.flatData(data=data)
        prop = {}

        for concepts in set(flat_data.index):
            activation_parameter = Transform.calculateProportions(data=flat_data, 
                                            conceptPair=concepts, nExperts=nExperts) 
            prop[concepts] = activation_parameter

        entropy_concept = {}
        for concept in prop.keys():
            p = prop[concept].values()
            res = -1*sum([i*np.log2(i) for i in p if i != 0])
            res = abs(res)
            entropy_concept[concept] = res
        
        # Prepare a formated dataframe
        entropy_concept = {k:[v] for k,v in entropy_concept.items()}
        entropy_concept = pd.DataFrame(entropy_concept).T
        entropy_concept.index.set_names(['From','To'], inplace=True)
        entropy_concept.columns = ['Entropy']
        entropy_concept = entropy_concept.sort_index(level=[0,1])

        return entropy_concept


from abc import ABC, abstractmethod
from fcmpy.expert_fcm.input_validator import type_check


class GetMethod(ABC):
    """
        Get methods from a store.
    """ 
    @abstractmethod
    def get(**kwargs):
        raise NotImplementedError('Get method is not defined!')


# Expert-based FCMs
class ReaderStore(GetMethod):
    """
        Methods of reading data files.
    """
    from fcmpy.expert_fcm.reader import XLSX
    from fcmpy.expert_fcm.reader import CSV
    from fcmpy.expert_fcm.reader import  JSON

    __methods = {'csv' : CSV, 'xlsx' : XLSX, 'json' : JSON}

    @staticmethod
    @type_check
    def get(method:str):
        if method in ReaderStore.__methods.keys():
            return ReaderStore.__methods[method]
        else:
            raise ValueError('The reader method is not defined.')


class EntropyStore(GetMethod):
    """
        Methods of calculating entropy.
    """
    from fcmpy.expert_fcm.entropy import InformationEntropy

    __methods = {'entropy' : InformationEntropy}

    @staticmethod
    @type_check
    def get(method:str):
        if method in EntropyStore.__methods.keys():
            return EntropyStore.__methods[method]
        else:
            raise ValueError('The enrtopy method is not defined.')


class MembershipStore(GetMethod):
    """
        Methods of generating membership functions.
    """
    from fcmpy.expert_fcm.membership import TriangularMembership
    from fcmpy.expert_fcm.membership import GaussianMembership
    from fcmpy.expert_fcm.membership import TrapezoidalMembership

    __methods = {'trimf' : TriangularMembership, 'gaussmf': GaussianMembership, 'trapmf' : TrapezoidalMembership}

    @staticmethod
    @type_check
    def get(method:str):
        if method in MembershipStore.__methods.keys():
            return MembershipStore.__methods[method]
        else:
            raise ValueError('The membership method is not defined.')


class ImplicationStore(GetMethod):
    """
        Fuzzy implication rules.
    """
    from fcmpy.expert_fcm.implication import Mamdani
    from fcmpy.expert_fcm.implication import Larsen

    __methods = {'Mamdani' : Mamdani, 'Larsen' : Larsen}

    @staticmethod
    @type_check
    def get(method:str):
        if method in ImplicationStore.__methods.keys():
            return ImplicationStore.__methods[method]
        else:
            raise ValueError('The implication method is not defined.')


class AggregationStore(GetMethod):
    """
        Fuzzy aggregation rules.
    """
    from fcmpy.expert_fcm.aggregation import Fmax
    from fcmpy.expert_fcm.aggregation import AlgSum
    from fcmpy.expert_fcm.aggregation import EinsteinSum
    from fcmpy.expert_fcm.aggregation import HamacherSum

    __methods = {'fMax' : Fmax, 'algSum' : AlgSum, 'eSum' : EinsteinSum, 'hSum' : HamacherSum}

    @staticmethod
    @type_check
    def get(method:str):
        if method in AggregationStore.__methods.keys():
            return AggregationStore.__methods[method]
        else:
            raise ValueError('The aggregation method is not defined.')


class DefuzzStore(GetMethod):
    """
        Defuzzification methods.
    """
    from fcmpy.expert_fcm.defuzz import Centroid
    from fcmpy.expert_fcm.defuzz import Bisector
    from fcmpy.expert_fcm.defuzz import MeanOfMax
    from fcmpy.expert_fcm.defuzz import MinOfMax
    from fcmpy.expert_fcm.defuzz import MaxOfMax

    __methods = {'centroid' : Centroid, 'bisector' : Bisector, 'mom' : MeanOfMax, 'som' : MinOfMax, 'lom' : MaxOfMax}

    @staticmethod
    @type_check
    def get(method:str):
        if method in DefuzzStore.__methods.keys():
            return DefuzzStore.__methods[method]
        else:
            raise ValueError('The defuzzification method is not defined.')


# Simulator
class InferenceStore(GetMethod):
    """
        Methods of FCM inference.
    """
    from fcmpy.simulator.inference import Kosko
    from fcmpy.simulator.inference import ModifiedKosko
    from fcmpy.simulator.inference import Rescaled

    __methods = {'kosko' : Kosko, 'mKosko' : ModifiedKosko, 'rescaled' : Rescaled}
    
    @staticmethod
    @type_check
    def get(method:str):
        if method in InferenceStore.__methods.keys():
            return InferenceStore.__methods[method]
        else:
            raise ValueError('The ifnerence method is not defined.')


class TransferStore(GetMethod):
    """
        Methods of FCM Transfer.
    """
    from fcmpy.simulator.transfer import Sigmoid
    from fcmpy.simulator.transfer import Bivalent
    from fcmpy.simulator.transfer import Trivalent
    from fcmpy.simulator.transfer import HyperbolicTangent
    __methods = {'sigmoid' : Sigmoid, 'bivalent' : Bivalent, 'trivalent' : Trivalent, 'tanh' : HyperbolicTangent}
    
    @staticmethod
    @type_check
    def get(method:str):
        if method in TransferStore.__methods.keys():
            return TransferStore.__methods[method]
        else:
            raise ValueError('The transfer method is not defined.')


class ConvergenceStore(GetMethod):
    """
        Methods for checking the convergence.
    """
    from fcmpy.simulator.convergence import AbsDifference

    __methods = {'absDiff' : AbsDifference}

    @staticmethod
    @type_check
    def get(method:str):
        if method in ConvergenceStore.__methods.keys():
            return ConvergenceStore.__methods[method]
        else:
            raise ValueError('The convergence method is not defined.')


# Intervention
class InterventionStore(GetMethod):
    """
        Methods of FCM Interventions.
    """
    from fcmpy.intervention.interventionConstructor import SingleShot
    from fcmpy.intervention.interventionConstructor import Continuous
    __methods = {'single_shot' : SingleShot, 'continuous' : Continuous}

    @staticmethod
    @type_check
    def get(method:str):
        if method in InterventionStore.__methods.keys():
            return InterventionStore.__methods[method]
        else:
            raise ValueError('The intervention type is not defined.')


# RCGA
class InitializationStore(GetMethod):
    """
        Initialization methods for the RCGA.
    """
    from fcmpy.ml.genetic.initialization import UniformInitialize

    __methods = {'uniform' : UniformInitialize}

    @staticmethod
    @type_check
    def get(method:str):
        if method in InitializationStore.__methods.keys():
            return InitializationStore.__methods[method]
        else:
            raise ValueError('The initialization method is not defined.')


# Normalization for RCGA
class NormalizationStore(GetMethod):  
    """
        Methods of normalizing the RCGA fitness function.
    """
    from fcmpy.ml.genetic.normalization import NT
    from fcmpy.ml.genetic.normalization import T
    __methods = {'L1' : NT, 'L2' : NT, 'LInf' : T}

    @staticmethod
    @type_check
    def get(method:str):
        if method in NormalizationStore.__methods.keys():
            return NormalizationStore.__methods[method]
        else:
            raise ValueError('The normalization method is not defined.')


# Auxilary methods for RCGA
class AuxiliaryStore(GetMethod):
    """
        Auxiliary functions for the RCGA fitness function.
    """
    from fcmpy.ml.genetic.auxiliary import H

    __methods = {'h' : H}

    @staticmethod
    @type_check
    def get(method:str):
        if method in AuxiliaryStore.__methods.keys():
            return AuxiliaryStore.__methods[method]
        else:
            raise ValueError('The auxiliary function is not defined.')


# Error functions for RCGA
class MatrixErrorStore(GetMethod):
    """
        Methods for calculating matrix error for the RCGA fitness function.
    """
    from fcmpy.ml.genetic.matrix_error import StachError

    __methods = {'stach_error' : StachError}

    @staticmethod
    @type_check
    def get(method:str):
        if method in MatrixErrorStore.__methods.keys():
            return MatrixErrorStore.__methods[method]
        else:
            raise ValueError('The error function is not defined.')


# Fitness functions for RCGA
class FitnessStore(GetMethod):
    """
        Methods for calculating fitness.
    """
    from fcmpy.ml.genetic.fitness import StachFitness

    __methods = {'stach_fitness' : StachFitness}

    @staticmethod
    @type_check
    def get(method:str):
        if method in FitnessStore.__methods.keys():
            return FitnessStore.__methods[method]
        else:
            raise ValueError('The fitness function is not defined.')


# Selection functions for RCGA
class SelectionStore(GetMethod):
    """
        Methods of selection for RCGA.
    """
    from fcmpy.ml.genetic.selection import Tournament
    from fcmpy.ml.genetic.selection import RouletteWheel


    __methods = {'tournament' : Tournament, 'roulette':RouletteWheel}

    @staticmethod
    @type_check
    def get(method:str):
        if method in SelectionStore.__methods.keys():
            return SelectionStore.__methods[method]
        else:
            raise ValueError('The select function is not defined.')


# Recombination functions for RCGA
class RecombinationStore(GetMethod):
    """
        Methods of selection for RCGA.
    """
    from fcmpy.ml.genetic.recombination import OnePointCrossover
    from fcmpy.ml.genetic.recombination import TwoPointCrossover


    __methods = {'one_point_crossover' : OnePointCrossover, 'two_point_crossover':TwoPointCrossover}

    @staticmethod
    @type_check
    def get(method:str):
        if method in RecombinationStore.__methods.keys():
            return RecombinationStore.__methods[method]
        else:
            raise ValueError('The recombine function is not defined.')


# Mutation functions for RCGA
class MutationStore(GetMethod):
    """
        Methods of mutation operations for RCGA.
    """
    from fcmpy.ml.genetic.mutation import RandomMutation
    from fcmpy.ml.genetic.mutation import NonUniformMutation


    __methods = {'random' : RandomMutation, 'non_uniform':NonUniformMutation}

    @staticmethod
    @type_check
    def get(method:str):
        if method in MutationStore.__methods.keys():
            return MutationStore.__methods[method]
        else:
            raise ValueError('The mutate function is not defined.')


# Replacement functions for RCGA
class ReplacementStore(GetMethod):
    """
        Methods of replacing candidate solutions for steady state RCGA.
    """
    from fcmpy.ml.genetic.replacement import CdrwReplacement

    __methods = {'CRDW': CdrwReplacement}

    @staticmethod
    @type_check
    def get(method:str):
        if method in ReplacementStore.__methods.keys():
            return ReplacementStore.__methods[method]
        else:
            raise ValueError('The replace function is not defined.')



import numpy as np
import pandas as pd 
import re
import collections
import json
from abc import ABC, abstractmethod
from fcmpy.expert_fcm.input_validator import type_check
from fcmpy.expert_fcm.checkers import ConsistencyCheck
from fcmpy.expert_fcm.checkers import ColumnsCheck


class ReadData(ABC):
    """
        Class of methods for reading in data.
    """
    @abstractmethod
    def read() -> collections.OrderedDict:
        raise NotImplementedError('read method is not defined!')


class CSV(ReadData):
    """
        Read data from a .csv file.
    """
    def __init__(self):
        pass

    @type_check
    def __conceptParser(self, string: str, sepConcept: str) -> dict:
        """
            Parse the csv file column names. Extract the antecedent -> 
            consequent pairs and the polarity of the causal relationship.

            Parameters
            ----------
            string: str
                    the column name that need to be parsed
            
            sepConcept: str
                        the separation symbol (e.g., '->') that separates 
                        the antecedent from the consequent in the columns of a csv file
            
            Return
            -------
            y: dict
                keys --> antecedent, consequent, polarity
        """
        dt = {}
        pattern = f'[a-zA-Z]+.+->.+.(\(\+\)|\(\-\))' # detect the column head pattern: antesedent -> consequent (polarity)
        patterMatch = bool(re.search(pattern, string))
        
        if patterMatch:
            dt['polarity'] = re.search(r'\((.*?)\)', string).group(1)
            concepts = string.split(sepConcept)
            dt['antecedent'] = re.sub(r'\([^)]*\)', '', concepts[0]).strip() 
            dt['consequent'] = re.sub(r'\([^)]*\)', '', concepts[1]).strip()
            return dt
        else:
            raise ValueError('The $antecedent$ $->$ $consequent (sign)$ \
                                format is not detected! Check the data format!')
    
    @type_check
    def __extractRowData(self, data: dict, sepConcept: str, 
                            linguisticTerms: list) -> pd.DataFrame:
        """
            Convert csv data fromat to a dataframe with columns representing the linguistic terms
            and rows representing the participants inputs (i.e., expert inputs).

            Parameters
            ----------
            data: dict
                    data to be extracted
            sepConcept: str
                        the separation symbol (e.g., '->') that separates the antecedent from the
                        consequent in the columns of a csv file
            
            linguisticTerms: list
                                list of linguistic terms
            
            Return
            -------
            y: pandas.DataFrame
        """
        ltValence = [i.lower().strip('+-') for i in linguisticTerms if i.startswith(('-', '+'), 0, 1)] # get the terms that have valence

        # Columns of the dfs (add the no causality term)
        dict_data = []
        for i in data.keys():
            _ = {i: 0 for i in linguisticTerms}
            conceptsParsed = self.__conceptParser(string=i, sepConcept=sepConcept)
            _['From'] = conceptsParsed['antecedent']
            _['To'] = conceptsParsed['consequent']

            # no causality cases
            if data[i].lower() in ltValence:
                if conceptsParsed['polarity'] == '+':
                    _[str('+'+data[i].lower())] = 1
                else:
                    _[str('-'+data[i].lower())] = 1
            else:
                _[data[i].lower()] = 1
            
            dict_data.append(_)
        
        return pd.DataFrame(dict_data).fillna(0)
    
    @type_check
    def read(self, **kwargs) -> collections.OrderedDict:
        """ 
            Read data from a csv file.

            Other Parameters
            ----------------
            **filepath : str

            **linguisticTerms: dictionary
                                dictionary of linguistic terms used to express causality between concepts.

            **sepConcept: str
                        the separation symbol (e.g., '->') that separates the antecedent from the consequent in the columns of a csv file

            **csvSep: str,
                        separator of the csv file (read more in pandas.read_csv)

            Return
            -------
            data: collections.OrderedDict
                    ordered dictionary with the formatted data.
        """
        filePath = kwargs['filePath']
        linguisticTerms = kwargs['linguisticTerms']
        linguisticTerms = [i.lower() for i in list(linguisticTerms.keys())] # Create a list

        try:
            sepConcept = kwargs['params']['sep_concept']
        except:
            sepConcept = '->'
        
        try:
            csvSep = kwargs['params']['csv_sep']
        except: # set the default csv sep to ','
            csvSep = ','

        data = pd.read_csv(filePath, sep=csvSep)
        dataOd = collections.OrderedDict()
        for i in range(len(data)):
            _ = data.iloc[i].to_dict()
            expertData = self.__extractRowData(data=_, sepConcept=sepConcept, linguisticTerms=linguisticTerms)
            dataOd[f'Expert{i}'] = expertData
        
        data = dataOd

        ColumnsCheck.checkColumns(data=data) # check whether From -> To columns exist.

        return data


class XLSX(ReadData):
    """
        Read data from .xlsx file.
    """
    @staticmethod
    @type_check
    def read(**kwargs) -> collections.OrderedDict:
        """ 
            Read data from .xlsx file.
            
            Other Parameters
            ----------------
            **filePath : str

            **checkConsistency: Bool
                                check the consistency of ratings across the experts.
                                default --> False
            
            **engine: str,
                        the engine for excel reader (read more in pd.read_excel)
                        default --> "openpyxl"
            
            Return
            -------
            data: collections.OrderedDict
                    ordered dictionary with the formatted data.
        """
        filePath = kwargs['filePath']

        try:
            checkConsistency = kwargs['params']['check_consistency']
        except:
            checkConsistency = False

        try:
            engine = kwargs['params']['engine']
        except:
            engine = 'openpyxl'

        data = pd.read_excel(filePath, sheet_name=None,  engine=engine)
        data = collections.OrderedDict(data)

        ColumnsCheck.checkColumns(data=data) # check whether From -> To columns exist.

        if checkConsistency:
            ConsistencyCheck.checkConsistency(data=data) # check the consistency of the linguistic ratings.
        
        return data


class JSON(ReadData):
    """
    Read data from .json file.
    """
    @staticmethod
    @type_check
    def read(**kwargs) -> collections.OrderedDict:
        """ 
            Read data from a .json file.

            Other Parameters
            ----------------
            **filePath : str, path object or file-like object

            **checkConsistency: Bool
                                check the consistency of ratings across the experts.
                                default --> False
            
            Return
            -------
            data: collections.OrderedDict
                    ordered dictionary with the formatted data.
        """
        filePath = kwargs['filePath']

        try:
            checkConsistency = kwargs['params']['check_consistency']
        except:
            checkConsistency = False

        f = open(filePath) 
        dataJson = json.load(f)
        f.close()
        d = {}
        for i in dataJson.keys():
            d[i] = dataJson[i]
        data = collections.OrderedDict([(i, pd.DataFrame(d[i]).replace(r'^\s*$', np.nan, regex=True)) for i in d])

        ColumnsCheck.checkColumns(data=data) # check whether From -> To columns exist.

        if checkConsistency:
            ConsistencyCheck.checkConsistency(data=data) # check the consistency of the linguistic ratings.

        return data


import unittest
import pandas as pd
from fcmpy.simulator.simulator import FcmSimulator

class TestSimulator(unittest.TestCase):
    
    def setUp(self):

        C1 = [0.0, 0.0, 0.6, 0.9, 0.0, 0.0, 0.0, 0.8]
        C2 = [0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.5]
        C3 = [0.0, 0.7, 0.0, 0.0, 0.9, 0.0, 0.4, 0.1]
        C4 = [0.4, 0.0, 0.0, 0.0, 0.0, 0.9, 0.0, 0.0]
        C5 = [0.0, 0.0, 0.0, 0.0, 0.0, -0.9, 0.0, 0.3]
        C6 = [-0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        C7 = [0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.4, 0.9]
        C8 =[0.1, 0.0, 0.0, 0.0, 0.0, 0.1, 0.6, 0.0]
        df = pd.DataFrame([C1,C2, C3, C4, C5, C6, C7, C8], 
                            columns=['C1','C2','C3','C4','C5','C6','C7','C8'])
        self.weight_matrix = pd.DataFrame([C1,C2, C3, C4, C5, C6, C7, C8], 
                        columns=['C1','C2','C3','C4','C5','C6','C7','C8'])

        self.init_state = {'C1': 1, 'C2': 1, 'C3': 0, 'C4': 0, 'C5': 0,
                        'C6': 0, 'C7': 0, 'C8': 0}
        self.sim = FcmSimulator()
        
    def test_simulation(self):
        res_k = self.sim.simulate(initial_state=self.init_state, weight_matrix=self.weight_matrix, transfer='sigmoid', inference='kosko', thresh=0.001, iterations=50, l=1)
        res_mK = self.sim.simulate(initial_state=self.init_state, weight_matrix=self.weight_matrix, transfer='sigmoid', inference='mKosko', thresh=0.001, iterations=50, l=1)
        res_r = self.sim.simulate(initial_state=self.init_state, weight_matrix=self.weight_matrix, transfer='sigmoid', inference='rescaled', thresh=0.001, iterations=50, l=1)

        eql_k = res_k.loc[len(res_k)-1] # equilibruim for Kosko's menthod
        eql_mK = res_mK.loc[len(res_mK)-1] # equilibruim for Kosko's menthod
        eql_r = res_r.loc[len(res_r)-1] # equilibruim for rescaled menthod

        # test the results against the one presented in the fcm inference package in R by Dikopoulou & Papageorgiou.
        equilibrium_mK = [0.7258851, 0.7907061, 0.7694508, 0.8124733, 0.8192938, 0.8399006, 0.9099403, 0.9557739]
        equilibrium_k = [0.5481291, 0.6004012, 0.5814756, 0.620872, 0.6279569, 0.6653852, 0.7617789, 0.8416584]
        equilibrium_r = [0.4998427, 0.4985174, 0.4989474, 0.4984212, 0.4980938, 0.4885541, 0.4855651, 0.4853266]

        # check if the initial equilibrium = to the one by Dikopoulou & Papageorgiou in R.
        self.assertEqual([round(i, 4) for i in eql_k], [round(i, 4) for i in equilibrium_k])
        self.assertEqual([round(i, 4) for i in eql_mK], [round(i, 4) for i in equilibrium_mK])
        self.assertEqual([round(i, 4) for i in eql_r], [round(i, 4) for i in equilibrium_r])

    def test_simulation_bi(self):
        res_mK = self.sim.simulate(initial_state=self.init_state, weight_matrix=self.weight_matrix, transfer='bivalent', inference='mKosko', thresh=0.001, iterations=50, l=1)
        self.assertEqual(len(set(res_mK.values.flatten())), 2)
        self.assertEqual(max(res_mK.values.flatten()), 1)
        self.assertEqual(min(res_mK.values.flatten()), 0)

    def test_simulation_tri(self):
        init_state = self.init_state.copy()
        init_state['C1'] = -1
        res_mK = self.sim.simulate(initial_state=init_state, weight_matrix=self.weight_matrix, transfer='trivalent', inference='mKosko', thresh=0.001, iterations=50, l=1)
        self.assertEqual(len(set(res_mK.values.flatten())), 3)
        self.assertEqual(max(res_mK.values.flatten()), 1)
        self.assertEqual(min(res_mK.values.flatten()), -1)

    def test_simulation_tanh(self):
        init_state = self.init_state.copy()
        init_state['C1'] = -1
        res_mK = self.sim.simulate(initial_state=init_state, weight_matrix=self.weight_matrix, transfer='tanh', inference='mKosko', thresh=0.001, iterations=50, l=1)
        self.assertEqual(max(res_mK.values.flatten()), 1)
        self.assertEqual(min(res_mK.values.flatten()), -1)

    def test_stableConcepts(self):
        self.weight_matrix['C1'] = 0
        res_k = self.sim.simulate(initial_state=self.init_state, weight_matrix=self.weight_matrix, transfer='sigmoid', inference='kosko', thresh=0.001, iterations=50, l=1)
        self.assertEqual(len(set(res_k['C1'])), 1)

    def test_outputConcepts(self):
        out = ['C4', 'C1']
        res_mK = self.sim.simulate(initial_state=self.init_state, weight_matrix=self.weight_matrix, transfer='sigmoid', inference='mKosko', thresh=0.001, iterations=50, l=1, output_concepts=out)
        _ = res_mK[out]
        residual = max(abs(_.loc[len(_)-1] - _.loc[len(_) - 2]))
        self.assertLessEqual(residual, 0.001)
        self.assertEqual([0.812650, 0.726125], list(_.loc[len(_)-1].values.round(6)))

if __name__ == '__main__':
    unittest.main()


import sys, os
import unittest
from fcmpy.expert_fcm.expert_based_fcm import ExpertFcm
import itertools
import collections
import pandas as pd 
import xlrd
import warnings
import functools
import numpy as np

class TestDataProcessor(unittest.TestCase):

    def setUp(self):
        self.fcm = ExpertFcm()
        self.nExpert = 6
        self.activation_param = {'M': 0.16, 'H': 0.5, 'VH': 0.33}
        self.fcm.universe = np.arange(-1, 1.001, 0.001)

        self.fcm.linguistic_terms = {
                                    '-VH': [-1, -1, -0.75],
                                    '-H': [-1, -0.75, -0.50],
                                    '-M': [-0.75, -0.5, -0.25], 
                                    '-L': [-0.5, -0.25, 0],
                                    '-VL': [-0.25, 0, 0],
                                    'NA': [-0.001, 0, 0.001],
                                    '+VL': [0, 0, 0.25],
                                    '+L': [0, 0.25, 0.50],
                                    '+M': [0.25, 0.5, 0.75],
                                    '+H': [0.5, 0.75, 1],
                                    '+VH':  [0.75, 1, 1]
                                    }
        
        self.fcm.fuzzy_membership = self.fcm.automf(method='trimf')

    def test_readExcel(self):
        data = self.fcm.read_data(file_path=os.path.abspath('unittests/test_cases/data_test.xlsx'), check_consistency=False, engine='openpyxl') 
        self.assertIsInstance(data, collections.OrderedDict)
        self.assertEqual(len(data), self.nExpert)

    def test_readExcelUnsure(self):
        data = self.fcm.read_data(file_path=os.path.abspath('unittests/test_cases/data_test_unsure.xlsx'), check_consistency=False, engine='openpyxl')
        self.assertIsInstance(data, collections.OrderedDict)
        self.assertEqual(len(data), self.nExpert)

    def test_readJson(self):
        data = self.fcm.read_data(file_path=os.path.abspath('unittests/test_cases/data_test.json'), check_consistency=False)
        self.assertIsInstance(data, collections.OrderedDict)
        self.assertEqual(len(data), 2)

    def test_readJsonUnsure(self):
        data = self.fcm.read_data(file_path=os.path.abspath('unittests/test_cases/data_test_unsure.json'), check_consistency=False)
        self.assertIsInstance(data, collections.OrderedDict)
        self.assertEqual(len(data), 2)

    def test_readCsv(self):
        self.fcm.linguistic_terms = {
                                    '-VH':  [-1, -1, -0.75],
                                    '-H':  [-1, -0.75, -0.50],
                                    '-M': [-0.75, -0.5, -0.25], 
                                    '-L': [-0.5, -0.25, 0],
                                    '-VL': [-0.25, 0, 0],
                                    'No Causality': [-0.001, 0, 0.001],
                                    '+VL': [0, 0, 0.25],
                                    '+L': [0, 0.25, 0.50],
                                    '+M': [0.25, 0.5, 0.75],
                                    '+H': [0.5, 0.75, 1],
                                    '+VH':  [0.75, 1, 1]
                                    }
        data = self.fcm.read_data(os.path.abspath('unittests/test_cases/jss_data.csv'), sep_concept='->', csv_sep=';')
        self.assertIsInstance(data, collections.OrderedDict)
        self.assertEqual(len(data), 5)

    def test_entropy(self):
        data = self.fcm.read_data(os.path.abspath('unittests/test_cases/data_test_entropy.xlsx'))
        entropy = self.fcm.entropy(data=data, method = 'entropy')
        self.assertGreaterEqual(min(entropy['Entropy']), 0)
        self.assertEqual(max(entropy['Entropy']), 0)
        
    def test_entropyUnsure(self):
        data = self.fcm.read_data(os.path.abspath('unittests/test_cases/data_test_entropy_unsure.xlsx'))
        entropy = self.fcm.entropy(data=data, method = 'entropy')
        self.assertEqual(round(max(entropy['Entropy']), 2), 0.65)

    def test_automf(self):
        mf  = self.fcm.automf(method = 'trimf')
        self.assertEqual(type(mf), dict)
        self.assertEqual(len(mf.keys()), len(self.fcm.linguistic_terms))

        # Check whether the values of the MF are within the fuzzy range 0,1.
        res_max = max(list(itertools.chain(*[list(mf[i]) for i in mf])))
        res_min = min(list(itertools.chain(*[list(mf[i]) for i in mf])))
        self.assertLessEqual(res_max, 1)
        self.assertGreaterEqual(res_min, 0)

        # check whether the generated universe is in the range -1,1 (default)
        self.assertLessEqual(round(max(self.fcm.universe), 5), 1)
        self.assertGreaterEqual(round(min(self.fcm.universe), 5), -1)

    def test_activate(self):
        mf  = self.fcm.fuzzy_membership
        act_M = self.fcm.fuzzy_implication(mf['+m'], weight=0.16, method ='Larsen')
        act_H = self.fcm.fuzzy_implication(mf['+h'], weight=0.5, method ='Larsen')
        act_VH = self.fcm.fuzzy_implication(mf['+vh'], weight= 0.33, method ='Larsen')
        activated = {'+m' : act_M, '+h' : act_H, '+vh' : act_VH}

        # Checks whether the activated membership functions are within the fuzzy range 0,1.
        for i in activated:
            self.assertLessEqual(max(activated[i]), 1)
            self.assertGreaterEqual(min(activated[i]), 0)

    def test_aggregate(self):
        mf  = self.fcm.fuzzy_membership
        act_M = self.fcm.fuzzy_implication(mf['+m'], weight=0.16, method ='Larsen')
        act_H = self.fcm.fuzzy_implication(mf['+h'], weight=0.5, method ='Larsen')
        act_VH = self.fcm.fuzzy_implication(mf['+vh'], weight= 0.33, method ='Larsen')
        activated = {'+m' : act_M, '+h' : act_H, '+vh' : act_VH}

        aggr = functools.reduce(lambda x,y: np.fmax(x,y), [activated[i] for i in activated.keys()])
        
        # the length of the aggregated function should be equal to that of the universe.
        self.assertEqual(len(aggr), len(self.fcm.universe))
        
        # Checks whether the aggregated function is in the fuzzy range. 
        self.assertLessEqual(max(aggr), 1)
        self.assertGreaterEqual(min(aggr), 0)

    def test_defuzzify(self):
        mf  = self.fcm.fuzzy_membership
        act_M = self.fcm.fuzzy_implication(mf['+m'], weight=0.16, method ='Larsen')
        act_H = self.fcm.fuzzy_implication(mf['+h'], weight=0.5, method ='Larsen')
        act_VH = self.fcm.fuzzy_implication(mf['+vh'], weight= 0.33, method ='Larsen')
        activated = {'+m' : act_M, '+h' : act_H, '+vh' : act_VH}

        aggr = functools.reduce(lambda x,y: np.fmax(x,y), [activated[i] for i in activated.keys()])

        dfuz = self.fcm.defuzz(x=self.fcm.universe, mfx=aggr)    
        self.assertAlmostEqual(dfuz, 0.72, 2)
    
    def test_build(self):
        self.fcm.fuzzy_membership = self.fcm.automf(method='trimf')
        
        data = self.fcm.read_data(file_path=os.path.abspath('unittests/test_cases/data_test.xlsx'), check_consistency=False, engine='openpyxl')
        weight_matrix = self.fcm.build(data=data, implication_method='Larsen')
        self.assertLessEqual(max(weight_matrix.max()), 1)
        self.assertGreaterEqual(min(weight_matrix.min()), -1)    

if __name__ == '__main__':
    unittest.main()


import numpy as np
from abc import ABC, abstractmethod  
from fcmpy.expert_fcm.input_validator import type_check

class Aggregate(ABC):
    """
        Fuzzy aggregation rule.
    """
    @abstractmethod
    def aggregate() -> np.ndarray:
        raise NotImplementedError('aggregate method is not defined.')


class Fmax(Aggregate):
    """
        Family max aggregation rule.
    """
    @staticmethod
    @type_check
    def aggregate(**kwargs) -> np.ndarray:
        """
            Family max aggregation rule.

            Other Parameters
            ----------
            **x, **y: numpy.ndarray,
                        "activated" membership functions of the linguistic terms
                            that need to be aggregated
            
            Return
            -------
            y: np.ndarray
                an aggregated membership function
        """
        x = kwargs['x']
        y = kwargs['y']

        return np.fmax(x, y)


class AlgSum(Aggregate):
    """
        Family Algebraic Sum.
    """
    @staticmethod
    @type_check
    def aggregate(**kwargs) -> np.ndarray:
        """
            Family Algebraic sum aggregation rule.

            Other Parameters
            ----------
            **x, **y: numpy.ndarray,
                        "activated" membership functions of the linguistic terms that need to be aggregated
            
            Return
            -------
            y: np.ndarray
                an aggregated membership function
        """
        x = kwargs['x']
        y = kwargs['y']

        return x+y - (x*y)


class EinsteinSum(Aggregate):
    """
        Family Einstein Sum.
    """
    @staticmethod
    @type_check
    def aggregate(**kwargs) -> np.ndarray:
        """
            Family Einstein sum aggregation rule.

            Other Parameters
            ----------
            **x, **y: numpy.ndarray,
                        "activated" membership functions of the linguistic terms that need to be aggregated
            
            Return
            -------
            y: float
                an aggregated membership function
        """
        x = kwargs['x']
        y = kwargs['y']

        return (x+y) / (1.0 + x*y)


class HamacherSum(Aggregate):
    """
        Family Hamacher Sum.
    """
    @staticmethod
    @type_check
    def aggregate(**kwargs) -> np.ndarray:
        """
            Family Hamacher sum aggregation rule.

            Other Parameters
            ----------------
            **x, **y: numpy.ndarray,
                        "activated" membership functions of the linguistic terms that need
                        to be aggregated
            
            Return
            -------
            y: np.ndarray
                an aggregated membership function
        """
        x = kwargs['x']
        y = kwargs['y']

        if x * y != 1.0:
            return (x+y - 2.0*x*y) / (1.0 - x*y)
        else:
            return 1.0


###########################################################################
##       Classes for evaluating the solutions of candidate solutions     ##
###########################################################################
import pandas as pd
from abc import ABC
from abc import abstractmethod
from fcmpy.ml.hebbian.update_state import FcmUpdate
from fcmpy.store.methodsStore import FitnessStore
from fcmpy.expert_fcm.input_validator import type_check


class Evaluation(ABC):
    """
        Interface for evaluation classes for RCGA.
    """
    @abstractmethod
    def evaluate(**kwargs):
        raise NotImplementedError('evaluate method is not defined.')


class PopulationEvaluation(Evaluation):
    """
        Evaluate the fitness of the population
    """
    @staticmethod
    @type_check
    def gen_data(state_vector:dict, weight_matrix:pd.DataFrame, transfer:str, 
                    inference:str, iterations:int, l:int, **kwargs) -> pd.DataFrame:
        """
            Update the state vector according to the selected inference method.

            Parameters
            ----------
            state_vector: dict
                            state vector of the concepts
                            keys ---> concepts, values ---> value of the associated concept

            weight_matrix: pd.DataFrame
                            N*N weight matrix of the FCM.

            transfer: str
                        transfer function to be applied on the updated concept values
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"

            inference: str
                        inference method for the FCM update
                        inference method --> "kosko", "mKosko", "rescaled"
            
            l: Union[float, int]
                    slop parameter for the sigmoid transfer function
                    default --> 0.98

            Return
            ------
            y: pd.DataFrame
                simulated data
        """
        sim = FcmUpdate()
        state_vector=state_vector
        res = [state_vector]
        for _ in range(iterations):
            s = sim.update(state_vector=state_vector, weight_matrix=weight_matrix, 
                            transfer=transfer, inference=inference, iterations=iterations, 
                            l=l, params=kwargs)
            res.append(s)
            state_vector=s
            
        return pd.DataFrame(res)

    @staticmethod
    def evaluate(**kwargs)->dict:
        """
            Evaluate the fitness of the population

            Parameters
            ----------
            population: dict
                        population of candidate chromosomes
            
            data: pd.DataFrame
                    longitudinal data (measurements of the concepts)
            
            inference: str
                        inference method for the FCM update
                        inference method --> "kosko", "mKosko", "rescaled"
                        default -> 'mKosko'
            
            transfer: str
                        transfer function to be applied on the updated concept values
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"
                        default --> 'sigmoid'
            
            l: int
                slop parameter for the sigmoid transfer function
                default --> 1
            
            fitness_type: str
                            type of fitness function to use
                            default -> 'stach_fitness'

            normalization_type: str
                                type of normalization to apply onto the matrix error
                                normalization type --> "L1", "L2", "LInf"
                                default --> L2
        
            Return
            ------
            y: dict
                population with the calculated fitness scores of each candidate solution
        """
        population = kwargs['population'].copy()
        data = kwargs['data'].copy()
        transfer = kwargs['transfer']
        inference = kwargs['inference']
        fitness_type = kwargs['fitness_type']
        normalization_type = kwargs['normalization_type']
        l = kwargs['l']
        a = kwargs['a']
        p = kwargs['p']
        
        # The first row of the data is the initial state
        initial_state = data.iloc[0].to_dict()
        iterations = len(data) - 1 # t-1 simulation steps

        # Select fitness method
        fitness = FitnessStore.get(method=fitness_type)
        
        # Evaluate the fitness of every candidate solution in the population
        for i in population.keys():
            # simulate a data with a length len(data)-1 based on the candidate solution
            res = PopulationEvaluation.gen_data(state_vector=initial_state, weight_matrix=pd.DataFrame(population[i]['solution'],
                                                    columns=initial_state.keys()), iterations=iterations, transfer=transfer, 
                                                    inference=inference, l=l, params=kwargs)
            # Calculate the fitness of the candidate solution based on the selected fitness method
            fit = fitness.calculate_fitness(data_simulated=res, data=data, normalization_type=normalization_type, p=p, a=a, params=kwargs)
            population[i]['fitness'] = fit
            
        return population


class ChromosomeEvaluation(Evaluation):
    """
        Evaluate the fitness of a single chromosome.
    """
    @staticmethod
    def evaluate(**kwargs)->dict:
        """
            Evaluate the fitness of a single chromosome.

            Parameters
            ----------
            child: dict
                    child chromosome to be evaluated
            
            data: pd.DataFrame
                    longitudinal data (measurements of the concepts)
            
            inference: str
                        inference method for the FCM update
                        inference method --> "kosko", "mKosko", "rescaled"
                        default -> 'mKosko'
            
            transfer: str
                        transfer function to be applied on the updated concept values
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"
                        default --> 'sigmoid'
            
            l: int
                slop parameter for the sigmoid transfer function
                default --> 1
            
            fitness_type: str
                            type of fitness function to use
                            default --> 'stach_fitness'

            normalization_type: str
                                type of normalization to apply onto the matrix error
                                normalization type --> "L1", "L2", "LInf"
                                default --> L2
        
            Return
            ------
            y: dict
                evaluated chromosome
        """
        child = kwargs['child'].copy()
        data = kwargs['data'].copy()
        transfer = kwargs['transfer']
        inference = kwargs['inference']
        fitness_type = kwargs['fitness_type']
        normalization_type = kwargs['normalization_type']
        l = kwargs['l']
        a = kwargs['a']
        p = kwargs['p']
        
        eval = PopulationEvaluation()
        # The first row of the data is the initial state
        initial_state = data.iloc[0].to_dict()
        iterations = len(data) - 1 # t-1 simulation steps

        # Select fitness method
        fitness = FitnessStore.get(method=fitness_type)

        res = eval.gen_data(state_vector=initial_state, weight_matrix=pd.DataFrame(child['solution'],
                                                    columns=initial_state.keys()), iterations=iterations, transfer=transfer, 
                                                    inference=inference, l=l, params=kwargs)
        
        fit = fitness.calculate_fitness(data_simulated=res, data=data, normalization_type=normalization_type, p=p, a=a, params=kwargs)
        return fit


###########################################################################
##            Classes for normalization of matrix error scores           ##
###########################################################################
from abc import ABC
from abc import abstractmethod
from fcmpy.expert_fcm.input_validator import type_check


class Normalization(ABC):
    """
        Interface for normalization classes for RCGA.
    """
    @abstractmethod
    def normalize(**kwargs):
        raise NotImplementedError('normalize method is not defined.')


class NT(Normalization):
    """
        Normalization class for Taxicab and Euclidean normalizations.
    """
    @staticmethod
    @type_check
    def normalize(**kwargs):
        """
            Normalization method for Taxicab and Euclidean normalizations (p = 1, and p e {1,2}).

            Parameters
            ----------
            n: int
                number of Concepts
            
            x: float
                matrix error
            
            t: int
                length of the data (i.e., time points of the longitudinal data)
            
            Return
            ------
            y: float
                normalized matrix error
        """
        x = kwargs['x']
        n = kwargs['n']
        t = kwargs['t']

        return x/(n*(t-1))


class T(Normalization):
    """
        Normalization class for maximum normalization.
    """
    def normalize(**kwargs):
        """
            Normalization method for maximum normalization (p=inf).

            Parameters
            ----------
            x: float
                matrix error
                
            t: int
                length of the data (i.e., time points of the longitudinal data)
            
            Return
            ------
            y: float
                normalized matrix error
        """
        x = kwargs['x']
        t = kwargs['t']
        
        return x/(t-1)


###########################################################################
##            Classes for recombination strategies                       ##
###########################################################################
import numpy as np
from abc import ABC
from abc import abstractmethod


class Recombination(ABC):
    """
        Interface for recombination classes for RCGA.
    """
    @abstractmethod
    def recombine(**kwargs):
        raise NotImplementedError('recombine method is not defined.')


class OnePointCrossover(Recombination):
    """
        One point crossover
    """
    def recombine(**kwargs)->tuple:
        """
            One-point crossover operation

            Parameters
            ----------
            parent1: np.ndarray
                        candidate solution
            
            parent2: np.ndarray
                        candidate solution
            
            Return
            ------
            y: tuple
                childOne, childTwo
        """
        parentOne = kwargs['parentOne']['solution'].copy().flatten()
        parentTwo = kwargs['parentTwo']['solution'].copy().flatten()

        childOne = {'solution': None, 'fitness' : 0}
        childTwo = {'solution': None, 'fitness' : 0}
        nConcepts = len(kwargs['parentOne']['solution'])

        # Determine the crossover point
        crossover_point = np.random.choice(range(nConcepts**2))

        childOne['solution'] = np.array(list(parentOne[:crossover_point]) + list(parentTwo[crossover_point:])).reshape(nConcepts, nConcepts)
        childTwo['solution'] = np.array(list(parentOne[crossover_point:]) + list(parentTwo[:crossover_point])).reshape(nConcepts, nConcepts)

        return childOne, childTwo
        

class TwoPointCrossover(Recombination):
    def recombine(**kwargs):
        # for future
        pass


from abc import ABC
from abc import abstractmethod
from random import random
import pandas as pd
import numpy as np
import tqdm
import os
import contextlib
from fcmpy.ml.hebbian.update_state import FcmUpdate

class Validation(ABC):
    @abstractmethod
    def validate(**kwargs):
        raise NotImplementedError('validate method is not defined!')


class HebbianValidate(Validation):
    def __init__(self, FcmSimulator):
        self.__sim = FcmSimulator()
        self.results = {}

    def __gen_state_vector(self, keys:list):
        """
            Randomly generate state vectors
            
            Parameters
            ----------
            keys: list
                    concepts of the FCM.
            
            Return
            ------
            y: dict
                keys ---> concepts, values ---> pseudo random numbers [0, 1]
        """
        return {k:random() for k in keys}

    def validate(self, n_validations:int, doc_values:dict, concepts:list,
                    weight_matrix:pd.DataFrame, transfer:str='sigmoid',
                    inference:str='mKosko', thresh:float=0.001, iterations:int=50, l:int=1, convergence = 'absDiff', **kwargs):
        """
            Validate the weight matrix by running FCM simulations with random initial conditions.

            Parameters
            ----------
            n_validations: int
                            number of validation steps

            doc_values: dict
                            Desired Output Concepts (DOCs) values.
                            keys ---> output concepts, values ---> desired output range ([min, max]).
                            e.g.,
                            doc_values = {'C1':[0.68,0.74], 'C5':[0.74,0.8]}

            concepts: list
                        list of concepts

            weight_matrix: pd.DataFrame,
                            N*N weight matrix of the FCM.

            transfer: str
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"

            inference: str
                        inference method --> "kosko", "mKosko", "rescaled"

            thresh: float
                        threshold for the error

            iterations: int
                            number of iterations

            convergence: str,
                            convergence method
                            default --> 'absDiff': absolute difference between the simulation steps
        """
        self.results = {c:[] for c in doc_values.keys()}

        for _ in tqdm.tqdm(range(n_validations)):
            init_state = self.__gen_state_vector(keys=concepts)

            # Redirect the print call in the simulation module
            with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
                res = self.__sim.simulate(initial_state=init_state, weight_matrix=weight_matrix, transfer=transfer,
                                            inference=inference, thresh=thresh, iterations=iterations, l=l, convergence=convergence, kwargs=kwargs).iloc[-1].to_dict()  
        
            for i in doc_values.keys():
                if doc_values[i][0] <= res[i] <= doc_values[i][1]:
                    for d in doc_values.keys():
                        self.results[d].append(res[d])
                else:
                    raise ValueError(f'The concept values are out of the desired bounds for the following initial conditions {init_state}')
            

class ISE(Validation):
    """
        In-Sample Error
    """
    def __init__(self):
        self.__sim = FcmUpdate()

    def validate(self, **kwargs):
        """
            In Sample Error

            Parameters
            ----------
            initial_state: dict
                            initial state of the concepts

            weight_matrix: pd.DataFrame
                            Candidate FCM solution (N*N weight matrix).

            transfer: str
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"

            inference: str
                        inference method --> "kosko", "mKosko", "rescaled"
            l: int
                parameter for the sigmoid function

            data: pd.DataFrame
                    longitudinal data
            
            Return
            ------
            y: float
                in sample error
        """
        init_state = kwargs['initial_state']
        weight_matrix = kwargs['weight_matrix']
        transfer = kwargs['transfer']
        inference = kwargs['inference']
        l = kwargs['l']
        data = kwargs['data']

        sim_data = [init_state]
        s_prior = init_state
        for _ in range(len(data)-1):
            res = self.__sim.update(state_vector=s_prior, weight_matrix=weight_matrix, transfer=transfer,
                                            inference=inference, l=l)
            s_prior = res
            sim_data.append(res)
        error = (1/((len(data)-1)*len(data.keys())))*(np.sum(np.sum(np.abs(pd.DataFrame(sim_data)-data))))
        return error


class OSE(Validation):
    """
        Out of Sample Error
    """
    def __init__(self):
        self.__sim = FcmUpdate

    def validate(self, **kwargs):
        """
            Out of Sample Error
        
            Parameters
            ----------
            weight_matrix: pd.DataFrame
                            Candidate FCM solution (N*N weight matrix).

            transfer: str
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"

            inference: str
                        inference method --> "kosko", "mKosko", "rescaled"
            l: int
                parameter for the sigmoid function

            data: pd.DataFrame
                    longitudinal data
            
            low: float
                    the lower bound of the concept values
            
            high: float
                    higher bound of the concept values
            
            k_validation: int
                            number of samples to generate for the validation
            
            Return
            ------
            y: tuple
                out of sample error, standard deviation
        """

        weight_matrix = kwargs['weight_matrix']
        transfer = kwargs['transfer']
        inference = kwargs['inference']
        l = kwargs['l']
        data = kwargs['data']
        low = kwargs['low']
        high = kwargs['high']
        concepts = data.keys()
        k_validation = kwargs['k_validation']
        error = []
        sim_data = []
        for i in range(k_validation):
            init_state = {k:np.random.uniform(low=low, high=high) for k in concepts}
            res = [init_state]
            # Simulate data points
            for t in range(len(data)-1):
                _ = self.__sim.update(state_vector=init_state, weight_matrix=weight_matrix, transfer=transfer,
                                                inference=inference, l=l)
                res.append(_)
            res = pd.DataFrame(res)
            if len(sim_data) > 1:
                error.append(np.abs(sim_data[1:]-res[1:]))
            
            sim_data = res
        std = np.std(error)

        avg_error=(1/(k_validation*(len(data)-1)*len(data.keys())))*np.sum(error)
        
        return avg_error, std
        


"""
New Idea: Inverse learning and topology post-optimization in Long-term Cognitive Network
reference:Deterministic learning of hybrid Fuzzy Cognitive Maps and network
reduction approaches
Gonzalo Nápoles a,b,∗, Agnieszka Jastrzębska c, Carlos Mosquera a, Koen Vanhoof a,
Władysław Homenda c

"""
from scipy.io import arff
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import mean_squared_error
from datetime import datetime
import numpy as np
import pandas as pd
import argparse
import os
import csv
import sys
import gc


def read_arff(f, **kwargs):
    '''
    reading from arff file
    :param f: filename
    :param kwargs: dictionary of paramters
    :return: X,Y and labels (classes)
    '''
    data, meta = arff.loadarff(f)
    df = pd.DataFrame(data)  # dataset
    class_att = meta.names()[-1]
    Y = df[class_att]

    # replacing labels
    labels = np.unique(Y)
    mapping = pd.Series([x[0] for x in enumerate(labels)], index=labels)
    Y = Y.map(mapping)  # target

    X = np.array(df.iloc[:, :-1], dtype=np.float64)

    return X, Y, labels


L, U = .1, .9


def logit(X, slope=1.0, h=0.0, q=1.0, v=1.0, **kwargs):
    return 1.0 / np.power(1.0 + q * np.exp(-slope * (X - h)), 1.0 / v)


def expit(Y, slope=1.0, h=0.0, q=1.0, v=1.0, **kwargs):
    return h - np.log((np.power(Y, -v) - 1.0) / q) / slope


class Model(object):
    '''
    FCM model
    '''
    def __init__(self, T=None, p=[], rule=0, b1=1.0, L=0, **kwargs):
        self.T, self.p = T, p
        self.weights = None
        self.p_out = p.copy()

        if rule == 0:
            self.rule = lambda a, W, out: logit(np.dot(a, W), *p)
        elif rule == 1:
            self.rule = lambda a, W, out: b1 * logit(np.dot(a, W), *p) + (1. - b1) * out[0]
        else:
            self.rule = lambda a, W, out: b1 * logit(np.dot(a, W), *p) + (1. - b1) * logit(
                W.diagonal() * np.sum(out[-L - 1:-1], axis=0), *p)

    def train(self, X_train, Y_train, verbose=False, callback=None, **kwargs):
        '''

        :param X_train: features
        :param Y_train: labels
        :param verbose:
        :param callback:
        :param kwargs: dict of params
        writes calculated weight matrix to the self.weights
        '''
        W = self.map(X_train)
        W[np.isnan(W)] = 0
        self.weights = W, np.random.random((X_train.shape[1], Y_train.shape[1]))

        _, out = self.test(X_train)

        # pseudo-inverse learning
        W_out = np.dot(np.linalg.pinv(out[-1]), expit(Y_train * (U - L) + L, *self.p))
        # normalizing the features importance[-1,1]
        mx = np.max(np.abs(W_out))
        if mx > 1:
            W_out /= mx
            self.p_out[0] *= mx
            self.p_out[1] /= mx
        # normalizing weights values importance[-1,1]
        mxW = np.max(np.abs(W))
        if mxW > 1:
            W /= mxW


        self.weights = W, W_out

    def test(self, X, **kwargs):
        '''
        testing obtained weight tmatrix
        :param X:
        :param kwargs: dict of params
        :return:
        '''
        W, W_out = self.weights
        out = [X]

        for t in range(1, self.T + 1):
            out.append(self.rule(out[t - 1], W, out))

        z = logit(np.dot(out[-1], W_out), *self.p)
        return (z - L) / (U - L), out

    def map(self, X_train, **kwargs):
        '''
        returns Weight matrix
        :param X_train: features
        :param kwargs: dict of params
        :return:
        '''
        n, m = X_train.shape
        T1 = np.sum(X_train, axis=0)
        T3 = np.sum(X_train ** 2, axis=0)

        W = np.random.random((m, m))
        for i in range(0, m):
            for j in range(0, m):
                d = n * T3[i] - T1[i] ** 2
                if d != 0:
                    W[i, j] = (n * np.sum(X_train[:, i] * X_train[:, j]) - T1[i] * T1[j]) / d
        return W


def cross_val(f, folds=10, **kwargs):
    '''
    cross valudation (loss)
    :param f: file
    :param folds: number of fold (def 10)
    :param kwargs: dict of parameters
    :return:
    '''
    X, Y, labels = read_arff(f)
    mse = []
    n_outputs = kwargs["M"]  # output variables
    if X.min() < 0. or X.max() > 1.:
        print("Numerical values need to be normalized.")
        raise Exception

    if kwargs["T"] is None:
        kwargs["T"] = X.shape[1] - n_outputs  # n_inputs

    seed = kwargs.get('seed', 1)
    np.random.seed(seed)
    skf = StratifiedKFold(n_splits=folds, shuffle=True, random_state=np.random.RandomState(seed))

    train_error = 0
    test_error = 0

    start = datetime.now()
    for train_index, test_index in skf.split(X, Y):
        X_train, X_test, Y_train, Y_test = X[train_index], X[test_index], Y[train_index], Y[test_index]

        model = Model(**kwargs)
        model.train(X_train[:, :-n_outputs], X_train[:, -n_outputs:].reshape(-1, 1), **kwargs)

        X_hat, _ = model.test(X_train[:, :-n_outputs])  # training error
        X_true = X_train[:, -n_outputs:].reshape(-1, 1)

        fold_train_error = np.sum(np.power(X_true - X_hat, 2) / n_outputs) / len(X_true)
        train_error += fold_train_error / folds

        y_hat, _ = model.test(X_test[:, :-n_outputs])
        y_true = X_test[:, -n_outputs:]

        fold_test_error = np.sum(np.power(y_true - y_hat, 2) / n_outputs) / len(y_true)
        test_error += fold_test_error / folds

    elapsed = datetime.now() - start

    slope, h, _, _ = kwargs["p"]
    return {
        # arguments
        "b1": "%.2f" % kwargs["b1"], "L": kwargs["L"], "slope": "%.2f" % slope, "h": "%.2f" % h,

        # metrics
        "train_error": train_error,
        "test_error": test_error,
        "training_time": elapsed.total_seconds() / folds,
        "weights":model.weights[0],
        "importance":model.weights[1]
        
    }


def run(**params):
    
    '''
    runs the whole algorithm for you, the only thing we ask is the parameters. Most of them can be default
    :param params:
    params = {
    'L':0, # For reasoning rule 3", default=0
    'M':1, #Output variables", default=1
    'T':None, # FCM Iterations default=None
    'b1':1.0,  # type=float, default=1.0
    'folds':10, # Number of folds in a (Stratified)K-Fold"
     'output':'./output.csv', # Output csv file", default="./output.csv"
     'p':[1.0, 1.0, 1.0, 1.0],
     'rule':0, # Reasoning rule", choices=[0, 1, 2], default=0
     'sources':['irisnorm.arff'],
     'verbose':False # store_true, verbosity }

    :return:
    '''

    
    data_sources = [f for f in params['sources'] if f.endswith('.arff')]


    results = []
    with open("./output.csv", mode='w') as csv_file:
        writer = csv.writer(csv_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        headers = None
        MSE_hist = []
        for f in data_sources:
            print("Processing {}".format(f))
            # here the model is being created and called
            result = cross_val(f, **params) 
            print(result)
            result['filename'] = f
            results.append(result)
            if headers is None:
                headers = list(result.keys())
                writer.writerow(["name"] + headers)

            MSE_hist += [result["test_error"]]
            writer.writerow([os.path.basename(f)[:-5]] + [result[k] for k in headers])
            
            if params['verbose']:
                print("MSE: %.4f" % MSE_hist[-1])

            csv_file.flush()
            sys.stdout.flush()
            gc.collect()

        print("MSE Average of the model across the %d datasets: %.4f" % (len(data_sources), np.average(MSE_hist)))
    return results 




import numpy as np
from abc import ABC, abstractmethod


class Convergence(ABC):
    """
        Class of FCM convergence methods
    """
    @abstractmethod
    def check_convergence() -> bool:
        raise NotImplementedError('Check_convergence method is not defined!')
    

class AbsDifference(Convergence):
    """
        Convergence check based on absolute difference .
    """
    @staticmethod
    def check_convergence(**kwargs) -> bool:
        """
            Compute the residuals (abs difference) of the outputConcepts
            between the simulation steps.
            
            Parameters
            ----------
            results: pd.DataFrame
                        the dataframe with the simulation results. 
                        Each row represents the results of a simulation step

            outputConcepts: None, list
                            if only specific outputConcepts should be considered
                            then a list of these concepts (in a string) should be passed.

            threshold: float, int
                        the threshold of the residuals to break the loop in the simulations.
            
            Return
            -------
            y: bool
                True if the residuals are <= the threshold, False if otherwise.
        """
        outputConcepts = kwargs['output_concepts']
        results = kwargs['results']
        threshold = kwargs['threshold']

        if outputConcepts:
            _ = results[outputConcepts]
            residual = max(abs(_.loc[len(results)-1] - _.loc[len(results) - 2]))
        else:
            residual = max(abs(results.loc[len(results)-1] - results.loc[len(results) - 2]))

        if residual <= threshold:
            return True
        else:
            return False


###########################################################################
##                    Classes for selection strategies                   ##
###########################################################################
import numpy as np
from abc import ABC
from abc import abstractmethod


class Selection(ABC):
    """
        Interface for selection classes for RCGA.
    """
    @abstractmethod
    def select(**kwargs):
         raise NotImplementedError('select method is not defined.')


class Tournament(Selection):
    """
        Tournament selection of candidate solutions
    """
    @staticmethod
    def select(**kwargs)->dict:
        """
            Tournament selection.

            Parameters
            ----------            
            n_participants: int
                            number of participants in the tournament
            
            population: int
                        population to be selected from
            
            Return
            ------
            y: dict
                selected chromosomes
        """
        n_participants = kwargs['n_participants']
        population = kwargs['population']
        selected_indexes = np.random.choice(list(population.keys()), size = n_participants)
        selected = {}

        s = sorted([population[i] for i in population if i in selected_indexes.flatten()], key = lambda i: i['fitness'], reverse=True)     
        selected[0] = s[0].copy()
        selected[1] = s[1].copy()

        return selected


class RouletteWheel(Selection):
    """
        Roulette wheel selection of candidate solutions
    """
    @staticmethod
    def select(**kwargs)->dict:
        """
            Roulette wheel selection

            Parameters
            ----------
            population: dict
                        population to be selected
            
            size: int
                    number of candidates (chromosomes) to be selected
            
            population: int
                        population to be selected from
            
            Return
            ------
            y: dict
                selected chromosomes 
        """
        population = kwargs['population']
        size = kwargs['size']
        f_score = [population[i]['fitness'] for i in population]
        total_f = np.sum(f_score)
        p = [i/total_f for i in f_score]

        selected_indexes = np.random.choice(list(population.keys()), size = size, p=p, replace=False)
        selected_solutions = {k:population.get(k,None).copy() for k in selected_indexes}

        return selected_solutions


###########################################################################
##                       Classes of fitness methods                      ##
###########################################################################
from abc import ABC
from abc import abstractmethod
from typing import Union
from fcmpy.store.methodsStore import AuxiliaryStore
from fcmpy.store.methodsStore import NormalizationStore
from fcmpy.store.methodsStore import MatrixErrorStore


class GaFit(ABC):
    """
        Interface for fitness classes for RCGA.
    """
    @abstractmethod
    def calculate_fitness(**kwargs):
        raise NotImplementedError('calculate_fitness method is not defined.')


class StachFitness(GaFit):
    """
        Fitness function based on Stach et al. 2005
    """
    @staticmethod
    def calculate_fitness(**kwargs) -> Union[float, int]:
        """
            Calculate the fitness of a candidate FCM (chromosome).

            Parameters
            ----------
            normalization_type: str
                                type of normalization function to be used

            data_simulated: pd.DataFrame
                                data generated based on the candidate FCM (chromosome)
            
            data: pd.DataFrame
                    real data
            
            p: int
                parameter for the matrix error

            a: int
                parameter for the auxiliary function

            Return
            ------
            y: float
                fitness score of a candidate solution
        """
        normalization_type = kwargs['normalization_type']
        data_simulated = kwargs['data_simulated']
        data = kwargs['data']
        a = kwargs['a']
        p= kwargs['p']
        
        error_type = 'stach_error'
        auxiliary_function_type = 'h'
        t = len(data)
        nConcepts = len(data.keys())

        # Get the methods for calculating the fitness function
        auxiliary = AuxiliaryStore.get(method=auxiliary_function_type)
        normalization = NormalizationStore.get(method=normalization_type)
        error = MatrixErrorStore.get(method=error_type)
        
        # Calculate the matrix error
        matError= error.calculate(data_simulated=data_simulated, data=data, p=p)
        # Normalize the calculated error
        normalized = normalization.normalize(x=matError, n=nConcepts, t=t)
        # Apply the auxiliary function on the normalized matrix error
        aux = auxiliary.f(x=normalized, a=a)

        return aux


import pandas as pd
import numpy as np
from typing import Union
from abc import ABC, abstractmethod
from fcmpy.store.methodsStore import InterventionStore
from fcmpy.expert_fcm.input_validator import type_check


class Intervention(ABC):
    """
        Test intervention scenarios.
    """
    @abstractmethod
    def add_intervention():
        raise NotImplementedError('add_intervention method is not defined!')

    @abstractmethod
    def remove_intervention():
        raise NotImplementedError('remove_intervention method is not defined!')

    @abstractmethod
    def test_intervention():
        raise NotImplementedError('test_intervention method is not defined!')


class FcmIntervention(Intervention):
    """
        The class includes methods for testing interventions (what-if scenarios) on top of a defined FCM structure.

        Methods:
            __init__(self, simulator)
            
            initialize(self, initial_state: dict, weight_matrix: Union[pd.DataFrame, np.ndarray], 
                                transfer: str, inference: str, thresh: float, iterations: int, l=1, 
                                output_concepts = None, convergence = 'absDiff',  **params)

            add_intervention(self, name, weights, effectiveness)

            remove_intervention(self, name)

            test_intervention(self, name, iterations = None)
    """
    def __init__(self, simulator):
        """
            Parameters
            ----------
            simulator: Simulator
        """
        self.__simulator = simulator()
        self.__interventions = {}
        self.__test_results = {}
        self.__equilibriums = {}
        self.__comparison_table = None

    @property
    def test_results(self):
        return self.__test_results
    
    @property
    def interventions(self):
        return self.__interventions
    
    @property
    def equilibriums(self):
        return pd.DataFrame(self.__equilibriums)
    
    @property
    def comparison_table(self):
        diff = {}
        df = pd.DataFrame(self.__equilibriums)
        for i in df.columns:
            diff[i] = ((df[i] - df.iloc[:, 0])/df.iloc[:, 0])*100
        self.__comparison_table = pd.DataFrame(diff)
        return self.__comparison_table

    @type_check
    def initialize(self, initial_state: dict, weight_matrix: Union[pd.DataFrame, np.ndarray], 
                            transfer: str, inference: str, thresh: float, iterations: int, l=1, 
                            output_concepts = None, convergence = 'absDiff',  **params):
        """
            Parameters
            ----------
            initial_state: dict
                            keys ---> concepts, values ---> initial states of the associated concepts

            weight_matrix: panda.DataFrame
                        causal weights between concepts

            transfer: str
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"

            inference: str
                        inference method --> "kosko", "mKosko", "rescaled"

            thresh: float
                        threshold for the error

            iterations: int
                            number of iterations

            l: 1
                A parameter that determines the steepness of the sigmoid function at values around 0.
            
            output_concepts: bool, list
                                the output concepts for the convergence check
                                default --> None
            
            convergence: str,
                            convergence method
                            default --> 'absDiff': absolute difference between the simulation steps

            **params: additional parameters
        """
        self.__weight_matrix = weight_matrix
        self.__initial_state=initial_state
        self.__transfer = transfer
        self.__inference = inference
        self.__thresh = thresh
        self.__iterations = iterations
        self.__l = l
        self.__output_concepts = output_concepts
        self.__convergence = convergence
        
        self.__test_results['baseline'] = self.__simulator.simulate(initial_state = self.__initial_state, weight_matrix = self.__weight_matrix,
                                                                transfer = self.__transfer, inference = self.__inference, thresh = self.__thresh, 
                                                                iterations = self.__iterations, l=self.__l, output_concepts = self.__output_concepts, convergence = self.__convergence, params = params)
        
        self.__equilibriums['baseline'] = self.test_results['baseline'].iloc[-1]

    @type_check
    def add_intervention(self, name, type='continuous', **kwargs):
        """
            Add an intervention node with the associated causal weights to the FCM.

            Parameters
            ----------
            name: str
                    name of the intervention
            
            type: str
                    type of intervention
                    default --> continuous

            impact: dict
                        keys --> concepts the intervention impacts, value: the associated causal weight

            effectiveness: float
                            the degree to which the intervention was delivered (should be between [0, 1])
                            default --> 1
        """
        if type != 'continuous':
            s = self.__initial_state.copy()
            s.update(kwargs['initial_state'])
            initial_state = s.copy()
        else:
            initial_state = self.__initial_state

        constructor = InterventionStore.get(type)()
        self.__interventions[name] = constructor.build(weight_matrix=self.__weight_matrix, initial_state=initial_state,
                                                        equilibriums = self.__equilibriums, params=kwargs)    
    
    @type_check
    def remove_intervention(self, name: str):
        """
            Remove intervention.

            Parameters
            ----------
            name: str
                    name of the intervention
        """
        del self.interventions[name]

    @type_check
    def test_intervention(self, name: str, iterations: int = None):
        """
            Test an intervention case.

            Parameters
            ----------
            name: str
                    name of the intervention
                    
            iterations: number of iterations for the FCM simulation
                            default ---> the iterations specified in the init.
        """
        if iterations:
            iterations = iterations
        else:
            iterations = self.__iterations

        weight_matrix = self.__interventions[name]['weight_matrix']
        state_vector = self.__interventions[name]['state_vector']
        
        self.__test_results[name] = self.__simulator.simulate(initial_state=state_vector, weight_matrix=weight_matrix, 
                                                                transfer=self.__transfer, inference=self.__inference, 
                                                                thresh=self.__thresh, iterations=iterations,
                                                                l = self.__l, output_concepts=self.__output_concepts,
                                                                convergence=self.__convergence)
        
        self.__equilibriums[name] = self.__test_results[name].iloc[-1][:len(self.__initial_state)]


import numpy as np
import pandas as pd
import collections
from datetime import date 
from tqdm import tqdm
from typing import Union
from fcmpy.expert_fcm.input_validator import type_check
from fcmpy.expert_fcm.transform import Transform


class ConsistencyCheck:
    """
        Class of methods for checking the consistency of the data.
    """
    @staticmethod
    def checkConsistency(data: Union[dict, collections.OrderedDict]):
        """
            Extract inconsistent ratings for the given linguistic terms in the
            supplied data. The method writes out an excel file with the
            inconsistencies and prints out a message if inconsistencies 
            were identified.
            
            Parameters
            ----------
            data : dict, collections.OrderedDict
        """
        current_date=date.today()
        flat_data = Transform.flatData(data=data)
        pairs = set(flat_data.index) # a set of all concept pairs.
        incon = {}

        for pair in tqdm(pairs):
            val = {}
            for expert in data.keys():
                dat = data[expert].copy(deep=True)
                dat.columns = [x.lower() for x in dat.columns] # set columns to lower case.
                dat = dat.set_index(['from', 'to']).replace(r'', np.nan) # replace the empty cells to np.nan
                dat[[i for i in dat if '-' in i]] = dat[[i for i in dat if '-' in i]] * -1
                v = dat.loc[pair].values[np.logical_not(np.isnan(dat.loc[pair].values))]
                if len(v) > 0:
                    val[expert] = int(v)
            if len(set(list(val.values()))) > 1:
                incon[pair] = val

        if incon:
            res = pd.DataFrame(incon).T
            res.index.set_names(['from', 'to'], inplace = True)
            res.to_excel(f'inconsistentRatings_{current_date.day}_{current_date.month}_{current_date.year}.xlsx', na_rep='NA')
            print(f'{list(res.index)} pairs of concepts were rated inconsistently across the experts. \
                 For more information check the inconsistentRatings_{current_date.day}_{current_date.month}_{current_date.year}.xlsx')


class ColumnsCheck:
    """
        Check the columns of the data.
    """
    @staticmethod
    @type_check
    def checkColumns(data: Union[dict, collections.OrderedDict]):
        """
            Checks whether the dataframe includes From --> To column. 
            It raises an error, if the columns are not found. 
            
            Parameters
            ----------
            data : dict
        """
        for expert in data.keys():
            if ('from' not in [x.lower() for x in data[expert].columns]) | \
                    ('to' not in [x.lower() for x in data[expert].columns]):
                raise ValueError('Columns From --> To were not found. Check the data!')


import numpy as np
from abc import ABC, abstractmethod


class Transfer(ABC):    
    """
        Class of FCM transfer methods.
    """ 
    @abstractmethod
    def transfer() -> np.array:
         raise NotImplementedError('Transfer method is not defined!')


class Sigmoid(Transfer):
    """
        Sigmoid transfer method
    """
    @staticmethod
    def transfer(**kwargs) -> np.array:
        """ 
            Sigmoid transfer function.
                
            Parameters
            ----------
            x : numpy.array,
                    the results of the FCM update function.
            l : int/float
                    A parameter that determines the steepness of the sigmoid function at values around 0. 
            
            Return
            -------
            y : numpy.array,
                    domain R,
                    range [0,1].
        """
        x = kwargs['x']
        l = kwargs['params']['l']
        e = np.exp(1)

        res = 1/(1+(e**(-l*x)))

        return res


class Bivalent(Transfer):
    """
        Bivalent transfer method
    """
    @staticmethod
    def transfer(**kwargs) -> np.array:
        """ 
            Bivalent transfer function.
                
            Parameters
            ----------
            x : numpy.array,
                    the results of the FCM update function.
            
            Return
            -------
            y : numpy.array,
                    domain R,
                    range [0;1].
        """
        x = kwargs['x']

        res = np.array([1 if i > 0 else 0 for i in x])

        return res


class Trivalent(Transfer):
    """
        Trivalent transfer function.
    """
    @staticmethod
    def transfer(**kwargs) -> np.array:
        """ 
            Trivalent transfer function.
                
            Parameters
            ----------
            x : numpy.array,
                    the results of the FCM update function.
            
            Return
            ----------
            y : numpy.array,
                    domain R,
                    range [-1,0,1].
        """
        x = kwargs['x']

        res = np.array([1 if i > 0 else -1 if i < 0 else 0 for i in x])

        return res


class HyperbolicTangent(Transfer):
    """
        Hyperbolic tangent transfer function.
    """
    @staticmethod
    def transfer(**kwargs) -> np.array:
        """ 
            Hyperbolic tangent transfer function.

            Parameters
            ----------
            x : numpy.array
                    the results of the FCM update function.
            
            Return
            -------
            y : numpy.array,
                    domain R,
                    range [-1,1].
        """
        x = kwargs['x']

        return np.tanh(x)



'''
For more information and details about the algorithm, please refer to
Pattern classification with Evolving Long-term Cognitive
Networks
Gonzalo Nápoles a,b,⇑, Agnieszka Jastrze˛bska c, Yamisleydi Salgueiro d
'''

import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn import datasets
from sklearn import model_selection
from sklearn.utils import shuffle
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import accuracy_score
from tensorflow.keras import backend as K
from tensorflow.python.keras.backend import set_session
from tensorflow.keras import regularizers
import matplotlib.pyplot as plt
from scipy.stats import entropy
from math import log, e
import os 
import warnings
warnings.filterwarnings("ignore")


class WeightRegularizer(tf.keras.regularizers.Regularizer):
    '''
    regularizing the wieghts
    '''

    def __init__(self, coef, mask):
        self.coef = coef
        self.mask = mask

    def __call__(self, weight_matrix):
        tensor = tf.convert_to_tensor(self.coef, np.float32)
        reg1 = K.sum(K.abs(tf.math.multiply(tf.math.subtract(weight_matrix,tensor), self.mask)))
        reg2 = 0.001 * K.sum(K.square(weight_matrix))
        return reg1 + reg2

    def get_config(self):
        return {'coefficients': self.coef, 'penalization': self.mask}
    

def run_model(file, layers=5, folds=5, epochs=1000, verbose=True, regularize=True):
    '''
    :param file: expects optimize .arff values with features vals [0,1]
    :param layers: numbers of nn layers (def 5)
    :param folds: how many validation folds (def 5)
    :param epochs: number of epochs (def 1000)
    :param verbose:
    :param regularize:
    :return: accuracy and weights matrix
    '''
    X, y, out = read_arff(file)
    hidden = len(X[0])
    
    skf = StratifiedKFold(n_splits=folds)
    skf.get_n_splits(X, y)
    
    acc_arr = []
    ent_arr = []
    for train_index, test_index in skf.split(X, y):
        X_train, X_test = np.asarray(X[train_index]).astype(np.float32),np.asarray(X[test_index]).astype(np.float32) #X[train_index], X[test_index]
        y_train, y_test = np.asarray(y[train_index]).astype(np.float32),np.asarray(y[test_index]).astype(np.float32) #y[train_index], y[test_index]
#         X = np.asarray(X).astype(np.float32)
        if(regularize):
            coef, mask = coefficients(matrix=X_train)
            network = [tf.keras.layers.Flatten()]     
    
        for i in range(layers):
            reg = None if (i == layers-1 or not regularize) else WeightRegularizer(coef, mask)
            dense = tf.keras.layers.Dense(hidden, activation=tf.nn.tanh, kernel_regularizer=reg)
            network.append(dense)

        network.append(tf.keras.layers.Dense(out, activation=tf.nn.softmax))

        model = tf.keras.models.Sequential(network)
        model.compile(optimizer='adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'])
        history = model.fit(X_train, y_train, epochs=epochs, verbose=0)
        
        if(verbose):
            plot_loss_weights(history, model, mask)
        
        weights = model.get_weights()
        relajo = error(weights, coef, mask)
        ent_arr.append(relajo)
        
        loss, acc = model.evaluate(X_test, y_test, verbose=0)
        acc_arr.append(acc)

    return np.mean(acc_arr), np.mean(ent_arr), weights 

def coefficients(matrix):
    n, m = matrix.shape
    temp1 = np.sum(matrix, axis=0)
    temp2 = np.sum(matrix**2, axis=0)
    
    df_data = pd.DataFrame(data=matrix,dtype=float)
    pearson = np.array(df_data.corr())
    
    mask = np.zeros((m, m))
    coef = np.zeros((m, m))
    for i in range(0, m):
        for j in range(0, m):
            den = n * temp2[i] - temp1[i] ** 2
            if (den != 0):
                coef[i,j] = (n * np.sum(matrix[:,i] * matrix[:,j]) - temp1[i] * temp1[j]) / den
                
                if(abs(pearson[i,j]) > 0.5):
                    mask[i,j] = log(n)
 
    return coef, mask

def plot_loss_weights(history, model, mask):
    '''
    plotting bar plots of individual weights for fully connected FCM
    :param history: data used to plot
    :param model: created model
    :param mask:
    :return:
    '''
    fig1, axes1 = plt.subplots(figsize=(15,5))
    fig2, axes2 = plt.subplots(figsize=(15,5))
        
    # plotting the training loss
    axes1.plot(history.history['loss'])
    axes1.set_title('model loss')
    axes1.set(ylabel='loss value', xlabel='epoch')

    # plotting the weights
    weights = model.get_weights()

    data = [xi.flatten().tolist() for xi in weights[:-2:2]]
    df = pd.DataFrame(data)
    
    header = []
    for i in range(len(mask)):
        for j in range(len(mask)):
            header.append('w'+str(i+1)+str(j+1)+"*" if (mask[i,j] != 0) else 'w'+str(i+1)+str(j+1))
            
    df.columns = header
    import seaborn as sns
    axes2 = sns.boxplot(data=df)
    axes2.set_title('weights')
    axes2.set(ylabel='value across layers', xlabel='weight')
    
def error(weights, coef, mask):
    '''
    returns error of the model for predicting correct classes
    :param weights: weight matrix
    :param coef: coefficients
    :param mask:
    :return: Error
    '''
    if(np.sum(mask) == 0):
        return 0
    
    coef = coef.flatten()
    mask = mask.flatten()
    array = []
    
    count = 0
    matrix = [xi.flatten() for xi in weights[:-2:2]]
    for i in range(len(matrix[0])):
        layer = column(matrix, i)
        for j in range(len(layer)):
            if(mask[j] > 0):
                den = max(1, max(layer[j], coef[j]))
                array.append(abs(layer[j] - coef[j]) / den)
                count += 1

    return np.sum(array)/count

def column(matrix, i):
    return [row[i] for row in matrix]
    
def read_arff(file):
    '''
    reading from .arff file
    consistency, in the MP algorithm we take the arff without header, so we cannot req header here
    :param file:
    :return:
    '''

    from scipy.io import arff
    data, meta = arff.loadarff(file)
    frame = pd.DataFrame(data)
    class_att = meta.names()[-1]
    y = frame[class_att]
    labels = np.unique(y)
    mapping = pd.Series([x[0] for x in enumerate(labels)], index = labels)
    y = np.array(y.map(mapping))
    X = np.array(frame)[:,0:-1]
    return X, y, len(labels)    
    
def run(path,folds=5):
    '''
    takes a directory where data file is (in .arff format) and do the whole calculation for you :)
    return fully connected FCM n x n for each fold and weight matrix n x nclasses
    :param path:
    :return:
    '''
  
    files = os.listdir(path)
    print(f"file in your data directory {files}. make sure they are .arff files!")
    print("running...")
    results = {}
    for file in files:
        if ".arff" not in file:
            print(f"{file} is not an .arff file")
        acc, ent, weights = run_model(path + "/" + file)
        # optimizing weights values for them to stay between [-1,1]
        for i in range(len(weights)):

            # print(np.abs(weights)[i])
            mx = np.max(np.abs(weights)[i])
            # print(f'max {mx}')
            if mx > 1:
                weights[i] /= mx
                # weights[i] *= mx
                # weights /= mx
        avgW = np.zeros((weights[0].shape[0],weights[0].shape[1]))     
        for i in np.arange(0,folds,step=2):
            avgW += weights[i]
        classW = np.zeros((weights[0].shape[0],weights[0].shape[1]))     
        for i in np.arange(1,folds,step=2):
            classW += weights[i]
        results[file] = {'acc':acc,'ent':ent,'weights':weights,'avgW':avgW/folds, 'classW':classW/folds}
        print(file.replace('.arff','') + "," + str(acc)+ "," + str(ent))
    return results




###########################################################################
##                   Classes for matrix error methods                    ##
###########################################################################
from abc import ABC
from abc import abstractmethod
import numpy as np


class MatrixError(ABC):
    """
        Interface for matrix error classes for RCGA.
    """
    @abstractmethod
    def calculate(**kwargs):
        raise NotImplementedError('calculate method is not defined.')


class StachError(MatrixError):
    """
        Calculate Matrix Error for RCGA based on Stach et al. proposal
    """
    @staticmethod
    def calculate(**kwargs):
        """
            Function for calculating matrix error for RCGA based on Stach et al. proposal

            Parameters
            ----------
            data_simulated: numpy.ndarray
                                data generated based on the candidate solution
            
            data: numpy.ndarray
                    training data
            
            p: int
                normalization parameter
            
            Return
            ------
            y: float, int
                Total matrix error
        """
        data_simulated = kwargs['data_simulated']
        data = kwargs['data']
        p = kwargs['p']
        
        return np.sum(np.sum(np.abs(data_simulated - data)**p))


import numpy as np
from abc import ABC
from abc import abstractmethod
from fcmpy.expert_fcm.input_validator import type_check
from fcmpy.store.methodsStore import InferenceStore
from fcmpy.store.methodsStore import TransferStore


class UpdateStateVector(ABC):
    
    @abstractmethod
    def update(**kwargs) -> np.array:
        raise NotImplementedError('Update method is not defined!')


class FcmUpdate(UpdateStateVector):
    """
        Update the state vector based on the selected inference method.
    """
    @staticmethod
    def update(**kwargs) -> dict:
        """
            Update the state vector according to the selected inference method.

            Parameters
            ----------
            state_vector: dict
                            state vector of the concepts
                            keys ---> concepts, values ---> value of the associated concept

            weight_matrix: pd.DataFrame
                            N*N weight matrix of the FCM.

            transfer: str
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"

            inference: str
                        inference method --> "kosko", "mKosko", "rescaled"
            
            Return
            ------
            y: dict
                keys -> concepts, values -> state values
        """
        weight_mat = kwargs['weight_matrix'].to_numpy()
        state_array = np.array(list(kwargs['state_vector'].values()))
        transfer_method = kwargs['transfer']
        inference_method = kwargs['inference']
        
        # get the methods for the simulation.
        inference = InferenceStore.get(inference_method)()
        transfer = TransferStore.get(transfer_method)()

        inferred = inference.infer(initial_state=state_array, weight_matrix=weight_mat, params=kwargs) # Inference
        state_vector = transfer.transfer(x=inferred, params=kwargs) # Apply transfer func on the results
        
        # convert to dict
        res = {i:y for i,y in zip(kwargs['weight_matrix'].columns, state_vector)}
        
        return res


class FcmUpdateAsynch(UpdateStateVector):
    """
        Asynchronous update of the FCM concept values.
    """
    @staticmethod
    @type_check
    def update(**kwargs) -> dict:
        """
            Asynchronously update the state vector.

            Parameters
            ----------
            source: str,
                        the activated concept

            state_vector: dict
                            state vector of the concepts
                            keys ---> concepts, values ---> value of the associated concept

            weight_matrix: pd.DataFrame
                            N*N weight matrix of the FCM.

            transfer: str
                        transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"            
            
            Return
            ------
            y: dict
                keys -> concepts, values -> state values
        """
        # extract the arguments
        source = kwargs['source']
        state = kwargs['state_vector'].copy()
        weight_matrix = kwargs['weight_matrix'].copy()

        # get the methods for the simulation.
        transfer_method = kwargs['transfer']
        transfer = TransferStore.get(transfer_method)()        
        
        for s in source:
            res = state[s] + weight_matrix[s].dot(list(state.values()))
            state[s] = transfer.transfer(x=res, params=kwargs)
        return state


###########################################################################
##                    Classes of replacement strategies                  ##
###########################################################################
import numpy as np
from scipy.spatial import distance
from abc import ABC
from abc import abstractmethod
import copy

class Replacement(ABC):
    """
        Interface for replacement classes
    """
    @abstractmethod
    def replace(**kwargs):
        raise NotImplementedError('replace method is not defined.')


class CdrwReplacement(Replacement):
    """
        Replacement based on contribution to diversity and replace worst strategy
    """
    @staticmethod
    def __diversity(chromosome1, chromosome2):
        """
            Calculate the diversity between a chromosome and the elements in the population based on Euclidean distance

            Parameters
            ----------
            chromosome: dict
                            chromosome whose diversity should be calculated

            population: dict
                        population of chromosomes

            Return
            --------
            y: float, 
                diversity score for a chromosome
        """
        divScore = distance.euclidean(chromosome1['solution'].flatten(), chromosome2['solution'].flatten())
        return divScore
        
    @staticmethod
    def __lessFitPop(chromosome, population):
        """
            Find a subpopulation whose elements have less fit than the passed chromosome.

            Parameters
            ----------
            chromosome: dict
                            child chromosome

            population: list of dict 
                            population of chromosomes

            Return
            --------
            y: list of dict 
                a subpopulation P
        """
        p = {}
        for i in population:
            if chromosome['fitness'] > population[i]['fitness']:
                p[i] = copy.deepcopy(population[i])
        lessFitPop = dict(sorted(p.items(), key=lambda k: k[1]['fitness']))
        
        return lessFitPop

    def replace(**kwargs):
        """
            Replacement based on contribution to diversity and replace worst.

            Parameters
            ----------
            child: dict
                        offspring 
                                
            population: dict
                        population of chromosomes

            Return
            --------
            y: list, 
                new population of chromosomes.
        """
        pop = kwargs['population']
        child = kwargs['child']
        # find the less fit chromosomes in the population for childOne and childTwo
        # if such subpopulation does not exist then the child does not survive
        childLessFitPop = CdrwReplacement.__lessFitPop(chromosome=child, population=pop)
        
        if childLessFitPop:
            childDiv = []
            for i in childLessFitPop:
                childDiv.append(CdrwReplacement.__diversity(chromosome1=child, chromosome2=childLessFitPop[i]))
            childDiv = min(childDiv)
            dvLessFitPop = {}
            for i in childLessFitPop:
                for y in pop:
                    if i != y:
                        dvLessFitPop[i] = CdrwReplacement.__diversity(chromosome1=childLessFitPop[i], chromosome2=pop[y])
            if childDiv > min(list(dvLessFitPop.values())):
                cmin = {k: v for k, v in dvLessFitPop.items() if v == min(dvLessFitPop.values())}
                pop[list(cmin.keys())[0]] = child
                return pop
            else:
                # replace worst
                pop[list(childLessFitPop.keys())[0]] = child
                return pop
        else: # if no replacement was done (i.e., child did not survive)
            return pop


from .expert_fcm.expert_based_fcm import ExpertFcm
from .simulator.simulator import FcmSimulator
from .intervention.intervention import FcmIntervention
from .ml.hebbian.nhl import NHL
from .ml.hebbian.ahl import AHL
from .ml.genetic.rcga import RCGA
from .ml.validation import HebbianValidate
from .ml.validation import ISE
from .ml.validation import OSE
from .ml import *


###########################################################################
##       Classes for auxiliary methods for the fitness scores            ##
###########################################################################
from abc import ABC
from abc import abstractmethod


class Auxiliary(ABC):
    """
        Interface for the auxiliary classes for the fitness score
    """
    @abstractmethod
    def f(**kwargs):
        raise NotImplementedError('f method is not defined.')


class H(Auxiliary):
    """
        Auxiliary classes for the fitness score
    """
    def f(**kwargs):
        """
            Auxiliary function for the fitness score

            Parameters
            ----------
            a: int
                a coefficient established experimentally
            
            x: float
                normalized matrix error
        
            Return
            ------
            y: float
                fitness score
        """
        a=kwargs['a']
        x=kwargs['x']
        return 1/(a*x+1)


import skfuzzy as fuzz
from abc import ABC, abstractclassmethod  
from fcmpy.expert_fcm.input_validator import type_check


class FuzzyMembership(ABC):
    """
        Fuzzy membership functions.
    """
    @abstractclassmethod
    def membershipFunction() -> dict:
        raise NotImplementedError('membershipFunction method is not defined')


class TriangularMembership(FuzzyMembership):
    """
        Triangular Fuzzy Membership Functions. 
    """
    @staticmethod
    @type_check
    def membershipFunction(**kwargs) -> dict:
        """
            Generate triangular membership functions.

            Other Parameters
            ----------------
            **linguisticTerms: dict,
                                terms and the associated parameters 
                                'abc' --> e.g., {'+VL': [0.25, 0.5, 0.75]}
            
            **universe: numpy.ndarray,
                        the universe of discourse                 

            Return
            -------
            y: dict,
                Generated membership functions. The keys are the 
                linguistic terms and the values are 1d arrays.
        """
        linguisticTerms = kwargs['linguistic_terms'] 
        universe = kwargs['universe'] 
        mfs = {}

        for term in linguisticTerms.keys():
            mfs[term] = fuzz.trimf(x = universe, abc=linguisticTerms[term])
        
        return mfs


class GaussianMembership(FuzzyMembership):
    """
        Gaussian Fuzzy Membership Functions. 
    """
    @staticmethod
    @type_check
    def membershipFunction(**kwargs) -> dict:
        """
            Generate Gaussian membership functions.

            Other Parameters
            ----------------
            **linguisticTerms: dict,
                                terms and the associated parameters 
                                'mean' and 'sigma' -> e.g., {'+VL': [0.25, 0.1]}}
            **universe: numpy.ndarray,
                        universe of discourse
                            
            Return
            -------
            y: dict,
                Generated membership functions. The keys are the linguistic terms 
                and the values are 1d arrays.
        """
        universe = kwargs['universe']
        linguisticTerms = kwargs['linguistic_terms'] 
        mfs = {}

        for term in linguisticTerms.keys():
            mfs[term] = fuzz.gaussmf(x = universe, mean=linguisticTerms[term][0],
                                        sigma = linguisticTerms[term][1])
        
        return mfs


class TrapezoidalMembership(FuzzyMembership):
    """
        Trapezoidal membership function.
    """
    @staticmethod
    @type_check
    def membershipFunction(**kwargs) -> dict:
        """
            Generate Trapezoidal membership functions.

            Other Parameters
            ----------------
            **linguisticTerms: dict,
                                terms and the associated parameters 'abcd' --> e.g., {'+VL': [0, 0.25, 0.5, 0.75]}
            **universe: numpy.ndarray,
                        universe of discourse

            Return
            -------
            y: dict,
                Generated membership functions. The keys are the linguistic terms and the values are 1d arrays.
        """
        linguisticTerms = kwargs['linguistic_terms'] 
        universe = kwargs['universe'] 
        mfs = {}
        for term in linguisticTerms.keys():
            assert linguisticTerms[term][0] <= linguisticTerms[term][1] <= linguisticTerms[term][2] <= linguisticTerms[term][3]
            mfs[term] = fuzz.trapmf(x = universe, abcd=linguisticTerms[term])
        
        return mfs


###########################################################################
##                    Implementation of the Steady State GA              ##
###########################################################################
import numpy as np
import pandas as pd
from tqdm import tqdm
from fcmpy.expert_fcm.input_validator import type_check
from fcmpy.ml.genetic.ga_interface import GA
from fcmpy.store.methodsStore import InitializationStore
from fcmpy.store.methodsStore import SelectionStore
from fcmpy.store.methodsStore import RecombinationStore
from fcmpy.store.methodsStore import MutationStore
from fcmpy.store.methodsStore import ReplacementStore
from fcmpy.ml.genetic.evaluation import ChromosomeEvaluation
from fcmpy.ml.genetic.evaluation import PopulationEvaluation
import copy


class SSGA(GA):
    """
        Steady State Real-Coded Algorithm for learning FCM connection matrices

        Parameters
        ----------
        data: pd.DataFrame
                longitudinal data
        
        init_type: str
                    type of initialization
                        default --> 'uniform'
        
        population_size: int
                            number of candidates to generate 
                            as an initial population
                            default --> 100
        
        fitness_type: str
                        fitness function
                        default --> 'stach_fitness'
        
        normalization_type: str
                            type of normalization to apply onto the matrix error
                            normalization type --> "L1", "L2", "LInf"
                            default --> L2
        a: int
            coefficient for the auxiliary function
            default -> 100

        p: int
            coefficient for the matrix error
            default -> 2 (for L2 norm.)
        
        inference: str
                    inference method for the FCM update
                    inference method --> "kosko", "mKosko", "rescaled"
                    default -> 'mKosko'
        
        transfer: str
                    transfer function to be applied on the updated concept values
                    transfer function --> "sigmoid", "bivalent", "trivalent", "tanh"
                    default --> "sigmoid"
        
        l: int
            slope parameter for the sigmoid transfer function
            default --> 1
    """
    @type_check
    def __init__(self, data:pd.DataFrame, init_type:str='uniform', population_size:int=100,
                    fitness_type:str='stach_fitness', normalization_type:str='L2', a:int=100, p:int=2,
                    inference:str='mKosko', transfer:str='sigmoid', l:int=1, **kwargs):
        
        self.__initialize = InitializationStore.get(init_type)
        self.__popSize = population_size
        self.__nConcepts = len(data.keys())
        self.__data = data
        __evaluation = PopulationEvaluation()
        self.__fitness_type = fitness_type
        self.__normalization_type = normalization_type
        self.__a = a
        self.__p = p
        self.__inference = inference
        self.__transfer = transfer
        self.__l=l

        # Step 1: Initialize the population
        population = self.__initialize.initialize(population_size = self.__popSize, 
                                                            n_concepts = self.__nConcepts, params=kwargs)
        # Evaluate the fitness of the candidate chromosomes in the population
        self.__population = __evaluation.evaluate(population=population, data=self.__data, l=self.__l, 
                                                            transfer=self.__transfer, inference=self.__inference,
                                                            fitness_type=self.__fitness_type, normalization_type=self.__normalization_type,
                                                            a=self.__a, p=self.__p)
    
    @type_check
    def run(self, n_iterations:int=100000, n_participants:int=5, recombination_type:str='one_point_crossover', replacement_strategy='CRDW',
                p_recombination:float=0.9, p_mutation:float=0.5, b:int=5, threshold:float=0.9, **kwargs):
        
        recombination = RecombinationStore.get(recombination_type)
        replacement = ReplacementStore.get(replacement_strategy)
        ev = ChromosomeEvaluation()
        old_population = copy.deepcopy(self.__population)
        best_candidate = {'solution': None, 'fitness':0}
        pbar = tqdm(range(n_iterations))
        
        for iteration in pbar:
            # Step 1: Selection
            select_method = np.random.choice(list(SelectionStore._SelectionStore__methods.keys()))
            selection = SelectionStore.get(method=select_method)
            selected = selection.select(population=old_population, size=2,
                                            n_participants=n_participants, params=kwargs)
            selected_indexes = [i for i in selected.keys()]

            # Step 2: Recombination
            change = np.random.choice([False, True], p=[1 - p_recombination, p_recombination]) 
            if change:
                parent1 = selected[selected_indexes[0]]
                parent2 = selected[selected_indexes[1]]
                childOne, childTwo = recombination.recombine(parentOne=copy.deepcopy(parent1), parentTwo=copy.deepcopy(parent2))
            else:
                childOne, childTwo = selected[selected_indexes[0]], selected[selected_indexes[1]]

            # Step 5: Mutation
            mutation_type =  np.random.choice(list(MutationStore._MutationStore__methods.keys())) 
            mutation = MutationStore.get(mutation_type)
            childOne = mutation.mutate(chromosome=copy.deepcopy(childOne), p_mutation=p_mutation, 
                                                        max_generations=n_iterations, nth_Iteration=iteration, b=b)
            childTwo = mutation.mutate(chromosome=copy.deepcopy(childTwo), p_mutation=p_mutation, 
                                                        max_generations=n_iterations, nth_Iteration=iteration, b=b)
            # Step 6: Evaluate the fitness of each solution
            childOne['fitness'] = ev.evaluate(child=childOne, data=self.__data, l=self.__l, 
                                            transfer=self.__transfer, inference=self.__inference,
                                            fitness_type=self.__fitness_type, normalization_type=self.__normalization_type,
                                            a=self.__a, p=self.__p).copy()
            childTwo['fitness'] = ev.evaluate(child=childTwo, data=self.__data, l=self.__l, 
                                            transfer=self.__transfer, inference=self.__inference,
                                            fitness_type=self.__fitness_type, normalization_type=self.__normalization_type,
                                            a=self.__a, p=self.__p).copy()
            # Step 7: Replace
            old_population = replacement.replace(child=copy.deepcopy(childOne), population=copy.deepcopy(old_population))
            old_population = replacement.replace(child=copy.deepcopy(childTwo), population=copy.deepcopy(old_population))
            
            # Get the best candidate of the current generation
            _ = copy.deepcopy(old_population[list(dict(sorted(old_population.items(), key=lambda k: k[1]['fitness'],reverse=True)).keys())[0]])
            # Update the best candidate if the best candidate of the current generation is better
            if _['fitness'] > best_candidate['fitness']:
                pbar.set_postfix({'fitness': _['fitness']})
            best_candidate = _
            # Step 8: Check termination
            if best_candidate['fitness'] >= threshold:
                print(f'RCGA identified a solution with fitness score <= {threshold}.')
                return best_candidate

        print(f'RCGA did not identify a solution with fitness score <= {threshold}. Returning a solution with the max fitness of the last generation.')
        return  best_candidate


###########################################################################
##                    Implementation of the Generationa GA               ##
###########################################################################
import numpy as np
import pandas as pd
from tqdm import tqdm
from fcmpy.expert_fcm.input_validator import type_check
from fcmpy.ml.genetic.ga_interface import GA
from fcmpy.store.methodsStore import InitializationStore
from fcmpy.store.methodsStore import SelectionStore
from fcmpy.store.methodsStore import RecombinationStore
from fcmpy.store.methodsStore import MutationStore
from fcmpy.ml.genetic.evaluation import PopulationEvaluation
import copy


class GRCGA(GA):
    """
        Generational Real-Coded Algorithm for learning FCM connection matrices

        Parameters
        ----------
        data: pd.DataFrame
                longitudinal data
        
        init_type: str
                    type of initialization
                    default --> 'uniform'
        
        population_size: int
                            number of candidates to generate 
                            as an initial population
                            default --> 100
        
        fitness_type: str
                        fitness function
                        default --> 'stach_fitness'
        
        normalization_type: str
                            type of normalization for the matrix error
                            default --> 'L2'

        a: int
            coefficient for the auxiliary function
            default --> 100

        p: int
            coefficient for the matrix error
            default --> 2 (for L2 norm.)
        
        inference: str
                    inference type for the FCM update function
                    default --> 'mKosko'
        
        tranfer: str
                 type of transfer function to be applied during the FCM update
                 default --> 'sigmoid'
        
        l: int
            slope parameter for the sigmoid function
            default --> 1
    """
    @type_check
    def __init__(self, data:pd.DataFrame, init_type:str='uniform', population_size:int=100,
                    fitness_type:str='stach_fitness', normalization_type:str='L2', a:int=100, p:int=2,
                    inference:str='mKosko', transfer:str='sigmoid', l:int=1, **kwargs):
        # Check the parameter ranges
        assert population_size % 2 == 0 # check if population size is even
        assert p > 0
        assert a > 0

        self.__initialize = InitializationStore.get(init_type)
        self.__popSize = population_size
        self.__nConcepts = len(data.keys())
        self.__data = data
        self.__evaluation = PopulationEvaluation()
        self.__fitness_type = fitness_type
        self.__normalization_type = normalization_type
        self.__a = a
        self.__p = p
        self.__inference = inference
        self.__transfer = transfer
        self.__l=l

        # Step 1: Initialize the population
        population = self.__initialize.initialize(population_size = self.__popSize, 
                                                            n_concepts = self.__nConcepts, params=kwargs)
        # Evaluate the fitness of the candidate chromosomes in the population
        self.__population = self.__evaluation.evaluate(population=population, data=self.__data, l=self.__l, 
                                                            transfer=self.__transfer, inference=self.__inference,
                                                            fitness_type=self.__fitness_type, normalization_type=self.__normalization_type,
                                                            a=self.__a, p=self.__p)

    @type_check
    def run(self, n_iterations:int=100000, n_participants:int=5, recombination_type:str='one_point_crossover',
                p_recombination:float=0.9, p_mutation:float=0.5, b:int=5, threshold:float=0.9, **kwargs):
        """
            Run the RCGA learning based on the generational approach

            Parameters
            ----------
            n_participants: int
                            number of participants in the turnament selection
                            deault -> 5
            
            p_mutation: float
                            probability of mutation
                            default -> 0.5
            
            p_recombination: float 
                                probability of recombination
                                default -> 0.9
            
            threshold: float
                        threshold of error (criteria for stopping the learning process)
                        default -> 0.999
            
            Return
            ------
            y: dict
                identified solution
        """
        # Check the parameter ranges
        assert n_iterations > 0
        assert n_participants > 0
        assert 0 <= p_recombination <= 1
        assert 0 <= p_mutation <= 1
        assert b > 0
        assert 0 <= threshold <= 1

        # Set up the recombination method
        recombination = RecombinationStore.get(recombination_type)
        # Create a copy of the initial population
        old_population = copy.deepcopy(self.__population)
        # Set up an obj. for the best candidate
        best_candidate = {'solution':None, 'fitness':0}
        pbar = tqdm(range(n_iterations))

        for iteration in pbar:
            new_pop = {}
            for _ in range(len(old_population)//2): # to maintain the population size
                # Step 1: Selection
                select_method = np.random.choice(list(SelectionStore._SelectionStore__methods.keys()))
                selection = SelectionStore.get(method=select_method)
                selected = selection.select(population=old_population, size=2,
                                                n_participants=n_participants, params=kwargs)
                selected_indexes = [i for i in selected.keys()]

                # Step 2: Recombination
                change = np.random.choice([False, True], p=[1 - p_recombination, p_recombination]) 
                if change:
                    parent1 = selected[selected_indexes[0]].copy()
                    parent2 = selected[selected_indexes[1]].copy()
                    childOne, childTwo = recombination.recombine(parentOne=parent1, parentTwo=parent2)
                else:
                    childOne, childTwo = selected[selected_indexes[0]], selected[selected_indexes[1]]

                # Step 5: Mutation
                mutation_type =  np.random.choice(list(MutationStore._MutationStore__methods.keys())) 
                mutation = MutationStore.get(mutation_type)
                new_pop[len(new_pop)] = mutation.mutate(chromosome=childOne, p_mutation=p_mutation, 
                                                            max_generations=n_iterations, nth_Iteration=iteration, b=b).copy()
                new_pop[len(new_pop)] = mutation.mutate(chromosome=childTwo, p_mutation=p_mutation, 
                                                            max_generations=n_iterations, nth_Iteration=iteration, b=b).copy()
            
            # Step 6: Re-evaluate the population
            evaluated = self.__evaluation.evaluate(population=new_pop, data=self.__data, l=self.__l, transfer=self.__transfer,
                                                        inference=self.__inference, fitness_type=self.__fitness_type, 
                                                        normalization_type=self.__normalization_type, a=self.__a, p=self.__p)

            old_population = evaluated

            # Get the best candidate of the current generation
            _ = old_population[list(dict(sorted(old_population.items(), key=lambda k: k[1]['fitness'],reverse=True)).keys())[0]]
            # Update the best candidate if the best candidate of the current generation is better
            if _['fitness'] > best_candidate['fitness']:
                best_candidate = _
                pbar.set_postfix({'fitness': _['fitness']})

            # Step 7: Check termination
            if best_candidate['fitness'] >= threshold:
                print(f'RCGA identified a solution with fitness score <= {threshold}.')
                return best_candidate
        
        print(f'RCGA did not identify a solution with fitness score <= {threshold}. Returning a solution with the max fitness throughout the search process.')
        return best_candidate


import unittest
import pandas as pd 
import numpy as np
from fcmpy import FcmSimulator
from fcmpy.store.methodsStore import AuxiliaryStore
from fcmpy.store.methodsStore import NormalizationStore
from fcmpy.store.methodsStore import MatrixErrorStore
from fcmpy.store.methodsStore import InitializationStore
from fcmpy.ml.genetic.evaluation import PopulationEvaluation

class TestRcga(unittest.TestCase):

    def setUp(self):
        w_init = np.asarray([[0,-0.4,-0.25,0,0.3],[0.36,0,0,0,0],
                        [0.45,0,0,0,0],[-0.9,0,0,0,0],
                        [0,0.6,0,0.3,0]])

        self.w_init_df = pd.DataFrame(w_init, columns=['C1', 'C2', 'C3', 'C4', 'C5'],
                                index = ['C1', 'C2', 'C3', 'C4', 'C5'])

        init_state = np.asarray([0.4,0.707,0.607,0.72,0.3])
        self.init_state_d = {'C1' : 0.40, 'C2': 0.7077, 'C3': 0.612, 'C4': 0.717, 'C5': 0.30}
        sim = FcmSimulator()
        self.data = sim.simulate(initial_state=self.init_state_d, weight_matrix=self.w_init_df,
                            transfer='sigmoid', inference='mKosko', thresh=0.001, iterations=50, l=1)

        self.auxiliary = AuxiliaryStore.get('h')
        self.normalization = NormalizationStore.get('L2')
        self.error = MatrixErrorStore.get('stach_error')

    def test_fitness(self):
        matError= self.error.calculate(data_simulated=self.data, data=(self.data), p=5)
        normalized = self.normalization.normalize(x=matError, n=5, t=6)
        aux = self.auxiliary.f(x=normalized, a = 10000)
        self.assertEqual(aux, 1)

    def test_UniformInitialization(self):
        initialize = InitializationStore.get('uniform')
        population = initialize.initialize(population_size = 100, 
                                                            n_concepts = 5)
        
        self.assertEqual(100, len(population))
    
    def test_PopulationEvaluation(self):
        initialize = InitializationStore.get('uniform')
        population = initialize.initialize(population_size = 100, 
                                                            n_concepts = 5)
        evaluation = PopulationEvaluation()
        init_state_d = self.data.iloc[0].to_dict()
        evaluated = evaluation.evaluate(population=population, data=self.data, l=5, transfer='sigmoid', inference='mKosko',
                                        normalization_type='L2', fitness_type='stach_fitness',
                                        a=10000, p=2)
        



import setuptools
import os 
def readme():
    with open("README.md", "r") as fh:
        return fh.read()

print(os.listdir())
with open('fcmpy.egg-info/requires.txt') as fid:
    INSTALL_REQUIRES = []
    for line in fid.readlines():
        if line == '' or line[0] == '#' or line[0].isspace():
            continue
        INSTALL_REQUIRES.append(line.strip())

setuptools.setup(
    name="fcmpy",
    version="0.0.1",
    author="Samvel Mkhitaryan, Philippe J. Giabbanelli, Maciej Wozniak, Nanne K. de Vries, Rik Crutzen",
    author_email="mkhitarian.samvel@gmail.com",
    description="Fuzzy Cognitive Maps for Behavior Change Interventions and Evaluation",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/SamvelMK/FcmBci.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8.1',
    install_requires=INSTALL_REQUIRES
)


###########################################################################
##                   Classes for mutation operations                     ##
###########################################################################
import numpy as np
from abc import ABC
from abc import abstractmethod
from fcmpy.expert_fcm.input_validator import type_check


class Mutation(ABC):
    """
        Interface for mutation classes for RCGA.
    """
    @abstractmethod
    def mutate(**kwargs):
        raise NotImplementedError('mutate method is not defined.')


class RandomMutation(Mutation):
    """
        Random uniform mutation based on Michalewicz, 1992.
    """
    def mutate(**kwargs)->dict:
        """
            Random uniform mutation based on Michalewicz, 1992.

            Parameters
            ----------
            chromosome: dict
                        the candidate FCM solution (chromosome). key --> solution,
                        fitness, value --> np.ndarray, float
            
            p_mutation: float
                            the mutation rate

            Return
            ------
            y: dict
                the mutated chromosome
        """
        chromosome = kwargs['chromosome'].copy()
        p_mutation = kwargs['p_mutation']

        for i in range(len(chromosome['solution'])):
            for j in range(len(chromosome['solution'])):
                change = np.random.choice([False, True], p=[1 - p_mutation, p_mutation]) # Decide whether the given gene should be mutated with p = p_mutation.
                if change:
                    chromosome['solution'][i,j] = np.random.uniform(-1,1)

        return chromosome


class NonUniformMutation(Mutation):
    """
        Non-uniform mutation based on Michalewicz, 1992.
    """
    @staticmethod
    @type_check
    def __nonUniCoef(t:int, T:int, y:float, b:int):
        """
            Calculate a coefficient for non uniform mutation

            Parameters
            ----------
            t: int
                current iteration

            T: int
                total number of iteration
            
            y: float

            b: int
                system parameter
            
            Return
            ------
            y: float
                a coefficient for the non uniform mutation (range [0,1])
        """
        r = np.random.uniform(0,1)
        mutationCoefficient = y*(1-r**((1-t/T)**b))

        return mutationCoefficient

    @staticmethod
    def mutate(**kwargs)->dict:
        """
            Non-uniform mutation based on Michalewicz, 1992.

            Parameters
            ----------
            chromosome: dict
                        the candidate FCM solution (chromosome). key -> solution,
                        fitness, value --> np.ndarray, float

            p_mutation: float
                            mutation rate

            max_generation: int
                            number of generations the RCGA should run

            n_iteration: int
                            the nth iteration of the RCGA

            b: int
                system parameter determining the degree of dependence
                of the mutation on the number of generation.

            Return
            ------
            y: dict
                the mutated chromosome
        """
        chromosome = kwargs['chromosome'].copy()
        p_mutation = kwargs['p_mutation']
        max_generations = kwargs['max_generations']
        nthIteration = kwargs['nth_Iteration']
        b = kwargs['b']

        for i in range(len(chromosome['solution'])):
            for j in range(len(chromosome['solution'])):
                change = np.random.choice([False, True], p=[1 - p_mutation, p_mutation]) # Decide whether the given gene should be mutated with p = p_mutation.
                if change:
                    m = np.random.randint(2, size=1)
                    if m > 0:
                        y = chromosome['solution'][i,j] + 1
                        # chromosome['solution'][i,j] = chromosome['solution'][i,j] - NonUniformMutation.__nonUniCoef(t=nthIteration, T=max_generations, y=y, b=b) 
                    else:
                        y = 1 - chromosome['solution'][i,j]
                        chromosome['solution'][i,j] = chromosome['solution'][i,j] + NonUniformMutation.__nonUniCoef(t=nthIteration, T=max_generations, y=y, b=b)

        return chromosome


from fcmpy.expert_fcm.input_validator import type_check
import numpy as np
from abc import ABC
from abc import abstractmethod

class UpdateParameters(ABC):
    @abstractmethod
    def update(**kwargs) -> float:
        raise NotImplementedError('update method is not defined.')


class Eta(UpdateParameters):
    """
        Update the learning parameter Eta.
    """
    @staticmethod
    @type_check
    def update(**kwargs) -> float:
        """
            Update learning parameter eta.
            
            Parameters
            ----------
            cycle: int,
                    learning cycle
            b1: float
                    default -> 0.02
            l1: float
                    default -> 0.1
        """
        b1 = kwargs['b1']
        l1 = kwargs['l1']
        cycle = kwargs['cycle']
        
        return b1 * np.exp(-l1*cycle)


class Gamma(UpdateParameters):
    """
        Update the decay parameter gamma.
    """
    @staticmethod
    @type_check
    def update(**kwargs):
        """
            Update learning parameter eta.
            
            Parameters
            ----------
            cycle: int,
                    learning cycle
            b2: float
                    default -> 0.04
            bl: float
                    default -> 1
        """
        b2 = kwargs['b2']
        l2 = kwargs['l2']
        cycle = kwargs['cycle']

        return b2*np.exp(-l2*cycle)


###########################################################################
##       Classes for initializing a population of candidate solutions    ##
###########################################################################
import numpy as np
from abc import ABC
from abc import abstractmethod


class Initialization(ABC):
    """
        Interface for initialization classes for RCGA.
    """
    @abstractmethod
    def initialize(**kwargs):
        raise NotImplementedError('initialize method is not defined.')


class UniformInitialize(Initialization):
    """
        Uniform generation of candidate solutions (chromosomes).
    """
    @staticmethod
    def initialize(**kwargs) -> dict:
        """
            Create a population of solutions (chromosomes) for the RCGA

            Parameters
            ----------
            population_size: int
                                number of solutions (population) to be generated
            n_concepts: int
                        number of geners in the chromosome (number of concepts in the FCM)
            
            Return
            ------
            y: dict
                dictionary keys ---> soluation (array of chromosome), fitness ---> float (fitness of a given chromosome)
        """
        population_size = kwargs['population_size']
        n_concepts = kwargs['n_concepts']
        generations = {}
        for i in range(population_size):
            generations[i] = {'solution': np.random.uniform(low=-1, high=1, size=(n_concepts, n_concepts)), 'fitness' : 0}
        return generations